"""
Constructor:

This file contains functions that can be called to construct a BigSMILES object.

There are two approaches:
* build it step by step;  where you open branch add atoms then close branch
* build it in chunks;  where you build the branch in a separate BigSMILES object and attach it.

"""
from __future__ import annotations
import logging

import bigsmiles.errors as errors
import bigsmiles.reference_data.chemical_data as chemical_data
from bigsmiles.data_structures.bigsmiles import Atom, Bond, BondDescriptor, Branch, StochasticFragment, \
    StochasticObject, BigSMILES, BondDescriptorAtom, has_node_attr, has_ring_attr, has_parent_attr
import bigsmiles.validation.validation_bigsmiles_obj as validation_bigsmiles_obj


## Utility functions for construction ## noqa
######################################################################################################################
def add_bond_to_connected_objects(bond: Bond):
    """
    Adds Bond to Atoms.bonds attribute
    also works on BondDescriptorAtom and stochastic objects
    """
    if bond.ring_id is not None and isinstance(bond.atom2, StochasticObject):
        atoms = reversed(bond)
    else:
        atoms = iter(bond)

    for i, obj in enumerate(atoms):
        if isinstance(obj, Atom):
            obj.bonds += [bond]
        elif isinstance(obj, BondDescriptorAtom):
            obj.bond = bond
        elif isinstance(obj, StochasticObject):
            if i == 0:
                obj.bond_right = bond
            else:
                obj.bond_left = bond


def add_bond_descriptor_to_stochastic_fragment(stoch_frag: StochasticFragment, loop_obj: Branch = None):
    """ Add bonding descriptor to stoch_frag.bonding_descriptors. Recursive. """
    if loop_obj is None:
        loop_obj = stoch_frag

    for obj in loop_obj.nodes:
        if isinstance(obj, BondDescriptorAtom) and obj.descriptor not in stoch_frag.bonding_descriptors:
            stoch_frag.bonding_descriptors.append(obj.descriptor)
        if isinstance(obj, Branch):
            add_bond_descriptor_to_stochastic_fragment(stoch_frag, obj)  # recursive


def get_common_bond(atom1: Atom, atom2: Atom | StochasticObject) -> Bond | None:
    """ Checks if the two atoms have a bond in common. """
    if isinstance(atom2, StochasticObject):
        if atom2.bond_right is not None and atom2.bond_right in atom1.bonds:
            raise errors.ConstructorError("An atom can't have two bonds to the same stochastic object.")
    else:
        for bond in atom1.bonds:
            if bond in atom2.bonds:
                return bond

    return None  # No common bond found


def remove_partial_ring(parent: has_ring_attr, ring_id: int):
    """ Given a ring_id, remove ring from parent. """
    for i, ring in enumerate(parent.rings):
        if ring.ring_id == ring_id:
            bond = parent.rings.pop(i)
            parent.bonds.remove(bond)
            if hasattr(parent, "parent"):
                parent.parent.bonds.remove(bond)
            break


def insert_obj_into_list(list_: list, obj, object_to_insert):
    """ Insert 'object_to_insert' in a list right behind 'obj'. """
    index_ = list_.index(obj) + 1
    for i, item in enumerate(list_[index_:]):
        if isinstance(item, Branch):  # put it at the end of existing branches
            continue

        list_.insert(index_ + i, object_to_insert)
        break


def get_prior(obj: has_node_attr, types_: tuple, flag: bool = False) -> Atom:
    """ Get prior """
    if obj.nodes:
        for node in reversed(obj.nodes):  # no atom can ever have more than 10 bonds
            if type(node) in types_:
                return node

        raise errors.ConstructorError("Trying to make a bond to a prior atom that isn't there.??")

    elif not flag:
        return get_prior(obj.parent, types_, flag=True)

    raise errors.ConstructorError("Bond attempted to be made to that has nothing to bond back to.")


def get_ring_parent(parent: has_parent_attr) -> BigSMILES | StochasticFragment:
    """ Recursively find a parent that has the 'ring' attribute. only BigSMILES and Stochastic Fragments do. """
    if hasattr(parent, "rings"):
        return parent

    return get_ring_parent(parent.parent)


def add_explict_hydrogen_to_stochastic_objects(bigsmiles_: BigSMILES):
    """ {[$][$]CC[$][$]} --> [H]{[$][$]CC[$][$]}[H] """
    if isinstance(bigsmiles_.nodes[0], StochasticObject) and not bigsmiles_.nodes[0].implicit_endgroups:
        if bigsmiles_.nodes[0].bd_left.bond_order > 1:
            raise errors.ConstructorError(
                f"Invalid BigSMILES. A double bond coming out of a stochastic object requires an "
                f"explict endgroup defined. \n Current Bigsmiles: {bigsmiles_}")

        insert_atom_and_bond(bigsmiles_, None, "", "H")

    if isinstance(bigsmiles_.nodes[-1], StochasticObject) and not bigsmiles_.nodes[-1].implicit_endgroups and \
            bigsmiles_.nodes[-1].bond_right is None:
        if bigsmiles_.nodes[-1].bd_right.bond_order > 1:
            raise errors.ConstructorError(
                f"Invalid BigSMILES. A double bond coming out of a stochastic object requires an "
                f"explict endgroup defined. \n Current Bigsmiles: {bigsmiles_}")

        add_bond_atom_pair(bigsmiles_, "", "H")


def exit_construction(parent: BigSMILES, syntax_fix: bool = True, validation: bool = True):
    """
    runs several methods to fix any syntax issues and runs more extensive validation

    Parameters
    ----------
    parent:
        BigSMILES to be checked
    syntax_fix:
        True to perform syntax fixes
    validation:
        True to run validation
    """
    if not isinstance(parent, BigSMILES):
        raise errors.ConstructorError(f"{type(parent)} is missing closing symbol.")

    if not parent:
        return parent
    if syntax_fix:
        add_explict_hydrogen_to_stochastic_objects(parent)
        validation_bigsmiles_obj.run_syntax_fixes(parent)
    if validation:
        validation_bigsmiles_obj.run_validation(parent)


## Functions for constructing a bigsmiles sequentially or step by step ## noqa
#######################################################################################################################
def add_atom(
        parent: has_node_attr,
        symbol: str,
        isotope: int | None = None,
        stereo: str | None = None,
        hydrogens: int | None = None,
        charge: int = 0,
        valence: int | None = None,
        class_: int | None = None,
        **kwargs
) -> has_node_attr:
    """
    appends an `Atom` to the end of the `parent`

    Parameters
    ----------
    parent:
    symbol:
        symbol symbol (e.g., H, C, O, Zn)
    isotope:
        isotope (e.g., [13C])
    stereo:
        stereochemistry [None, "@", "@@"] (None means not defined)
    hydrogens:
        number of explict hydrogens  (e.g., [CH2])
    charge:
        symbol charge  (e.g., [Fe+3])
    valence:
        The capacity to form bonds with other atoms
    class_:
        index of class (e.g., [C:1] class_ = 1)
    kwargs:
        any additional keyword arguments are accepted and set as additional attributes to the Atom

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    atom = Atom(parent._get_id(Atom), symbol, isotope, stereo, hydrogens, charge, valence, class_,
                parent=parent, **kwargs)
    parent.nodes.append(atom)
    parent.root.atoms.append(atom)

    return parent


def add_bond(
        parent: has_node_attr,
        bond_symbol: str | None,
        atom1: Atom | BondDescriptorAtom,
        atom2: Atom | BondDescriptorAtom | StochasticObject | None,
        **kwargs
) -> has_node_attr:
    """
    add 'Bond' between two 'Atoms' (typically not be used directly)

    Parameters
    ----------
    parent:
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    atom1:
        atom the bond starts at
    atom2:
        atom the bond ends at
    kwargs:
        any additional keyword arguments are accepted and set as additional attributes to the Bond

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    bond = Bond(parent._get_id(Bond), bond_symbol, atom1, atom2, parent=parent, **kwargs)
    parent.nodes.append(bond)
    parent.root.bonds.append(bond)
    add_bond_to_connected_objects(bond)

    return parent


def add_ring_by_index(parent: has_node_attr, ring_id: int, bond_symbol: str | None, **kwargs) -> has_node_attr:
    """
    add ring by index; rings are just 'Bond'

    !!! warning

        This function needs to be called twice to work; once at each time the parsing hits a ring index.

        'exit_construction()' will check to confirm that there is no uncompleted rings

    Parameters
    ----------
    parent:
    ring_id:
        index of ring
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    kwargs:
        any additional keyword arguments are accepted and set as additional attributes to the Bond

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    # check if ring_id already exists
    ring_parent = get_ring_parent(parent)
    for ring in ring_parent.rings:
        if ring.ring_id == ring_id:
            if ring.atom2 is not None:
                raise errors.ConstructorError(f"Ring already formed for ring id {ring_id}.")
            atom2 = get_prior(parent, (Atom, StochasticObject))

            # validation
            bond = get_common_bond(ring.atom1, atom2)
            if bond:
                logging.warning("Duplicate ring detected and merged into one with a higher bond order. ")
                bond.bond_order += chemical_data.bond_mapping[bond_symbol]
                remove_partial_ring(ring_parent, ring_id)
                return parent

            # Add atom2
            # ring.atom1, ring.atom2 = atom2, ring.atom1
            ring.atom2 = atom2

            # this is to handle that the multi-bond symbol (-, #) could be on either or both ring index
            if chemical_data.bond_mapping[bond_symbol] > ring.bond_order:
                ring.symbol = bond_symbol

            # aromatic_elements ring closure
            if ring.atom1.aromatic and ring.atom2.aromatic:
                ring.symbol = ":"

            add_bond_to_connected_objects(ring)
            return parent

    # Make new ring_id
    bond = Bond(parent._get_id(Bond), bond_symbol, get_prior(parent, (Atom, StochasticObject)), None, ring_id,
                parent=parent, **kwargs)
    parent.root.bonds.append(bond)
    ring_parent.rings.append(bond)

    return parent


def add_ring_from_atoms(parent: has_node_attr, atom1: Atom, atom2: Atom, bond_symbol: str | None = "", **kwargs) \
        -> has_node_attr:
    """
    add ring from with both 'Atoms'; rings are just 'Bond'

    Parameters
    ----------
    parent:
    atom1:
        atom the ring starts at
    atom2:
        atom the ring ends at
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    kwargs
        any additional keyword arguments are accepted and set as additional attributes to the Bond

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    ring_parent = get_ring_parent(parent)
    bond = get_common_bond(atom1, atom2)  # Check if ring already between two atoms
    if bond is None:
        # Make new ring_id
        bond = Bond(parent._get_id(Bond), bond_symbol, atom1, atom2, len(parent.root.rings) + 1,
                    parent=parent, **kwargs)
        parent.root.bonds.append(bond)
        ring_parent.rings.append(bond)
        add_bond_to_connected_objects(bond)
    else:
        # increase bond order of existing bond
        bond.bond_order += 1

    return parent


def _get_bonding_descriptor_atom(parent, descriptor: str, index_: int, bond_symbol: str | None, **kwargs) \
        -> BondDescriptorAtom:
    stoch_obj = _get_current_stochastic_object(parent)
    bd = _get_bonding_descriptor(stoch_obj, descriptor, index_, bond_symbol)
    return BondDescriptorAtom(parent._get_id(BondDescriptorAtom), bd, parent=parent, **kwargs)


def _get_bonding_descriptor(stoch_obj: StochasticObject, descriptor: str, index_: int, bond_symbol: str | None) \
        -> BondDescriptor:
    # check if bd in there already
    for bd in stoch_obj.bonding_descriptors:
        if descriptor == bd.descriptor and index_ == bd.index_:
            if bond_symbol is not None and bd.bond_symbol != bond_symbol:
                raise errors.ConstructorError("Multiple bond orders to same bonding descriptor.")
                # logging.warning("Multiple bond orders to same bonding descriptor.")
            return bd

    # create new bonding descriptor
    new_bd = BondDescriptor(stoch_obj, descriptor, index_, bond_symbol)
    stoch_obj.bonding_descriptors.append(new_bd)
    return new_bd


def _get_current_stochastic_object(parent) -> StochasticObject:
    """ Recursive"""
    if isinstance(parent, StochasticObject):
        return parent

    if hasattr(parent, "parent"):
        return _get_current_stochastic_object(parent.parent)
    else:
        raise errors.ConstructorError("No stochastic object found.")


def add_bonding_descriptor_atom(parent, descriptor: str, index_: int, bond_symbol: str | None = None, **kwargs) \
        -> has_node_attr:
    """
    appends 'BondDescriptorAtom' to the end of the `parent`

    Parameters
    ----------
    parent:
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    kwargs
        any additional keyword arguments are accepted and set as additional attributes to the BondDescriptorAtom

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_, bond_symbol, **kwargs)
    parent.nodes.append(bd_atom)
    parent.root.atoms.append(bd_atom)
    return parent


def add_bond_atom_pair(
        parent: has_node_attr,
        bond_symbol: str | None,
        symbol: str,
        isotope: int | None = None,
        stereo: str | None = None,
        hydrogens: int | None = None,
        charge: int = 0,
        valence: int = None,
        class_: int | None = None,
        kwargs_atom: dict = None,
        kwargs_bond: dict = None
) -> has_node_attr:
    """
    appends a `Bond` followed by an `Atom`

    Parameters
    ----------
    parent:
        parent
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    symbol:
        symbol symbol (e.g., H, C, O, Zn)
    isotope:
        isotope (e.g., [13C])
    stereo:
        stereochemistry [None, "@", "@@"] (None means not defined)
    hydrogens:
        number of explict hydrogens  (e.g., [CH2])
    charge:
        symbol charge  (e.g., [Fe+3])
    valence:
        The capacity to form bonds with other atoms
    class_:
        index of class (e.g., [C:1] class_ = 1)
    kwargs_atom:
        any additional keyword arguments are accepted and set as additional attributes to the Atom
    kwargs_bond
        any additional keyword arguments are accepted and set as additional attributes to the Bond

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    kwargs_atom = kwargs_atom if kwargs_atom is not None else {}
    kwargs_bond = kwargs_bond if kwargs_bond is not None else {}

    atom = Atom(parent._get_id(Atom), symbol, isotope, stereo, hydrogens, charge, valence, class_,
                parent=parent, **kwargs_atom)
    prior_atom = get_prior(parent, (Atom, BondDescriptorAtom, StochasticObject))
    add_bond(parent, bond_symbol, prior_atom, atom, **kwargs_bond)
    parent.nodes.append(atom)
    parent.root.atoms.append(atom)

    return parent


def add_bond_bonding_descriptor_pair(
        parent: StochasticFragment | Branch,
        bond_symbol: str | None,
        descriptor: str,
        index_: int,
        kwargs_bond_descriptor: dict = None,
        kwargs_bond: dict = None
) -> StochasticFragment | Branch:
    """
    appends a `Bond` followed by an `BondingDescriptorAtom`

    Parameters
    ----------
    parent:
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    kwargs_bond_descriptor:
        any additional keyword arguments are accepted and set as additional attributes to the BondDescriptorAtom
    kwargs_bond:
        any additional keyword arguments are accepted and set as additional attributes to the Bond

    Returns
    -------
    parent:
        returns the parent passed to it
    """
    kwargs_bond_descriptor = kwargs_bond_descriptor if kwargs_bond_descriptor is not None else {}
    kwargs_bond = kwargs_bond if kwargs_bond is not None else {}

    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_, bond_symbol, **kwargs_bond_descriptor)
    prior_atom = get_prior(parent, (Atom, BondDescriptorAtom, StochasticObject))
    add_bond(parent, bond_symbol, prior_atom, bd_atom, **kwargs_bond)
    parent.nodes.append(bd_atom)
    parent.root.atoms.append(bd_atom)
    return parent


def open_branch(parent: has_node_attr, **kwargs) -> Branch:
    """
    open branch

    !!! warning

        close_branch() must be called at some later point

    Parameters
    ----------
    parent:
    kwargs:
        any additional keyword arguments are accepted and set as additional attributes to the Branch

    Returns
    -------
    parent:
        returns Branch
    """
    if len(parent.nodes) == 0 and hasattr(parent, 'parent') and isinstance(parent.parent, Branch):
        raise errors.ConstructorError("BigSMILES string or branch can't start with branch")

    branch = Branch(parent._get_id(Branch), parent, **kwargs)
    parent.nodes.append(branch)

    return branch


def close_branch(parent: Branch) -> has_node_attr:
    """
    closes branch and appends it to `parent.parent`

    Parameters
    ----------
    parent:

    Returns
    -------
    parent:
        returns parent.parent

    """
    if not isinstance(parent, Branch):
        raise errors.ConstructorError("Error closing branch. Possible issues: "
                                      "\n\tClosing a branch with another intermediate node started."
                                      "\n\tNo starting branch symbol.")

    # check if branch is empty; if so remove it
    if len(parent.nodes) == 0:
        parent.parent.nodes.pop()

    return parent.parent


def open_stochastic_object(parent: has_node_attr, descriptor: str, index_: int, **kwargs) -> StochasticObject:
    """
    open `StochasticObject`

    !!! warning

        close_stochastic_object() must be called at a later point

    Parameters
    ----------
    parent:
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    kwargs
        any additional keyword arguments are accepted and set as additional attributes to the BondDescriptorAtom

    Returns
    -------
    parent:
        returns StochasticObject
    """
    stoch_obj = StochasticObject(parent._get_id(StochasticObject), parent, **kwargs)
    stoch_obj.bd_left = _get_bonding_descriptor(stoch_obj, descriptor, index_, None)
    parent.nodes.append(stoch_obj)

    return stoch_obj


def open_stochastic_object_fragment(parent: has_node_attr, descriptor: str, index_: int, **kwargs) \
        -> StochasticFragment:
    """
    opens `StochasticObject` and `StochasticFragment`

    !!! warning

        close_stochastic_fragment() must be called at a later point

        close_stochastic_object() must be called at a later point

    Parameters
    ----------
    parent:
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    kwargs
        any additional keyword arguments are accepted and set as additional attributes to the BondDescriptorAtom

    Returns
    -------
    parent:
        returns StochasticFragment
    """
    stoch_obj = open_stochastic_object(parent, descriptor, index_, **kwargs)

    return open_stochastic_fragment(stoch_obj)


def open_stochastic_object_fragment_with_bond(parent: has_node_attr, bond_symbol: str | None, descriptor: str,
                                              index_: int, **kwargs) -> StochasticFragment:
    """
    opens `StochasticObject` and `StochasticFragment`

    !!! warning

        close_stochastic_fragment() must be called at a later point

        close_stochastic_object() must be called at a later point

    Parameters
    ----------
    parent:
    bond_symbol:
        bond symbol ("", "=", "#", etc.)
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    kwargs
        any additional keyword arguments are accepted and set as additional attributes to the BondDescriptorAtom

    Returns
    -------
    parent:
        returns StochasticFragment
    """
    stoch_obj = StochasticObject(parent._get_id(StochasticObject), parent, **kwargs)
    stoch_obj.bd_left = _get_bonding_descriptor(stoch_obj, descriptor, index_, bond_symbol)

    # bond made without 'add_bond' function to ensure it's added to 'bond_left'
    prior_atom = get_prior(parent, (Atom, StochasticObject))
    bond = Bond(parent._get_id(Bond), bond_symbol, prior_atom, stoch_obj, parent=parent)
    parent.nodes.append(bond)
    parent.root.bonds.append(bond)
    add_bond_to_connected_objects(bond)
    parent.nodes.append(stoch_obj)

    return open_stochastic_fragment(stoch_obj)


def close_stochastic_object(parent: StochasticObject, descriptor: str, index_: int, bond_symbol: str | None) \
        -> has_node_attr:
    """
    closes `StochasticObject`

    Parameters
    ----------
    parent:
    descriptor:
        bonding descriptor symbol [<, >, $]
    index_:
        bonding descriptor index [1, inf]
    bond_symbol:
        bond symbol ("", "=", "#", etc.)

    Returns
    -------
    parent:
        returns parent.parent

    Raises
    ------
    ConstructorError
        Error closing StochasticObject. Possible issues:

        * Closing a StochasticObject with another intermediate node started.
        * No starting StochasticObject symbol.
    """
    if not isinstance(parent, StochasticObject):
        raise errors.ConstructorError("Error closing StochasticObject. Possible issues: "
                                      "\n\tClosing a StochasticObject with another intermediate node started."
                                      "\n\tNo starting StochasticObject symbol.")

    parent.bd_right = _get_bonding_descriptor(parent, descriptor, index_, bond_symbol)
    return parent.parent


def open_stochastic_fragment(parent: StochasticObject) -> StochasticFragment:
    """
    open `StochasticFragment`

    !!! warning

        close_stochastic_fragment() must be called at a later point


    Parameters
    ----------
    parent:

    Returns
    -------
    parent:
        returns StochasticFragment
    """
    stoch_frag = StochasticFragment(parent._get_id(StochasticFragment), parent)
    parent.nodes.append(stoch_frag)
    return stoch_frag


def close_open_stochastic_fragment(parent: StochasticFragment) -> StochasticFragment:
    """

    closes `StochasticFragment` and opens a new `StochasticFragment`

    Parameters
    ----------
    parent:

    Returns
    -------
    parent:
        returns a new StochasticFragment
    """
    if not isinstance(parent, StochasticFragment):
        raise errors.ConstructorError("Stochastic seperator can only follow a stochastic fragments.")

    parent = close_stochastic_fragment(parent)
    return open_stochastic_fragment(parent)


def close_stochastic_fragment(parent: StochasticFragment) -> StochasticObject:
    """

    Parameters
    ----------
    parent:

    Returns
    -------
    parent:
        returns parent.parent
    """
    if not isinstance(parent, StochasticFragment):
        raise errors.ConstructorError("Error StochasticFragment branch. Possible issues: "
                                      "\n\tClosing a branch with another intermediate node started.")

    add_bond_descriptor_to_stochastic_fragment(parent)

    # check for at-least one bonding descriptor
    if not parent.bonding_descriptors:
        raise errors.ConstructorError(f"No bonding descriptor in the stochastic fragment. ({repr(parent)})")

    return parent.parent


## functions for building BigSMILES in chunks ## noqa
###################################################################################################################
def append_bigsmiles_fragment(parent, bigsmiles_: BigSMILES, bond_symbol: str | None, **kwargs) -> BigSMILES:
    """

    Parameters
    ----------
    parent
    bigsmiles_
    bond_symbol
    kwargs

    Returns
    -------

    """
    if not isinstance(bigsmiles_.nodes[0], (Atom, StochasticObject)):
        raise errors.ConstructorError("First node must be an 'Atom' or 'StochasticObject' to added fragment.")
    if not parent:
        return bigsmiles_

    atom = bigsmiles_.nodes[0]
    prior_atom = get_prior(parent, (Atom, StochasticObject))
    add_bond(parent, bond_symbol, prior_atom, atom, **kwargs)

    set_new_parent(parent, bigsmiles_)

    # append fragment
    parent.nodes += bigsmiles_.nodes
    parent.root.atoms += bigsmiles_.atoms
    parent.root.bonds += bigsmiles_.bonds
    parent.root.rings += bigsmiles_.rings

    return parent


def attach_bigsmiles_branch(parent, bond_symbol: str | None | None, bigsmiles_: BigSMILES, index_: int,
                            kwargs_bond: dict = None, kwargs_branch: dict = None):
    """

    Parameters
    ----------
    parent
    bond_symbol
    bigsmiles_
    index_
    kwargs_bond
    kwargs_branch

    Returns
    -------

    """
    kwargs_bond = kwargs_bond if kwargs_bond is not None else {}
    kwargs_branch = kwargs_branch if kwargs_branch is not None else {}

    if index_ > len(parent.root.atoms) - 1:
        raise errors.ConstructorError(f"Branch can't be added. 'index' outside atom list."
                                      f"\n current smiles: {parent.root} \n desired atom index: {index_}")

    set_new_parent(parent, bigsmiles_)

    # make branch
    branch = Branch(parent._get_id(Branch), parent, **kwargs_branch)
    branch.nodes = bigsmiles_.nodes
    parent.root.atoms += bigsmiles_.atoms
    parent.root.bonds += bigsmiles_.bonds

    # make bond
    atom = parent.root.atoms[index_]
    bond = Bond(parent._get_id(Bond), bond_symbol, atom, branch.nodes[0], parent=parent, **kwargs_bond)
    branch.nodes.insert(0, bond)
    parent.root.bonds.append(bond)
    add_bond_to_connected_objects(bond)

    # insert branch (put it after all the other branches on the atom)
    insert_obj_into_list(parent.root.nodes, atom, branch)

    return parent


def insert_atom_and_bond(
        parent: has_node_attr,
        prior_atom: Atom | BondDescriptorAtom | Branch | StochasticObject | None,
        bond_symbol: str | None,
        element: str,
        isotope: int | None = None,
        stereo: str = '',
        hydrogens: int = 0,
        charge: int = 0,
        valance: int = None,
        class_: int | None = None,
        kwargs_atom: dict = None,
        kwargs_bond: dict = None
) -> has_node_attr:
    """

    Parameters
    ----------
    parent
    prior_atom
    bond_symbol
    element
    isotope
    stereo
    hydrogens
    charge
    valance
    class_
    kwargs_atom
    kwargs_bond

    Returns
    -------

    """
    kwargs_atom = kwargs_atom if kwargs_atom is not None else {}
    kwargs_bond = kwargs_bond if kwargs_bond is not None else {}

    atom = Atom(parent._get_id(Atom), element, isotope, stereo, hydrogens, charge, valance, class_,
                parent=parent, **kwargs_atom)

    if prior_atom is None and isinstance(parent, BigSMILES):  # add to beginning
        prior_atom = parent.nodes[0]
        bond = Bond(parent._get_id(Bond), bond_symbol, atom, prior_atom, parent=parent, **kwargs_bond)
        add_bond_to_connected_objects(bond)
        parent.nodes.insert(0, bond)
        parent.root.bonds.insert(0, bond)
        parent.nodes.insert(0, atom)
        parent.root.atoms.insert(0, atom)

    elif prior_atom == prior_atom.parent.nodes[-1]:  # add to end
        bond = Bond(parent._get_id(Bond), bond_symbol, parent.atoms[-1], atom, parent=parent, **kwargs_bond)
        add_bond_to_connected_objects(bond)
        parent.nodes.append(bond)
        parent.root.bonds.append(bond)
        parent.nodes.append(atom)
        parent.root.atoms.append(atom)

    validation_bigsmiles_obj.re_number_node_ids(parent.root)
    return parent


def insert_atom_into_bond(
        parent: has_node_attr,
        bond_to_insert_into: Bond,
        element: str,
        isotope: int | None = None,
        stereo: str = '',
        hydrogens: int = 0,
        charge: int = 0,
        valance: int = None,
        class_: int | None = None,
        kwargs_atom: dict = None,
) -> has_node_attr:
    """

    Parameters
    ----------
    parent
    bond_to_insert_into
    element
    isotope
    stereo
    hydrogens
    charge
    valance
    class_
    kwargs_atom

    Returns
    -------

    """
    kwargs_atom = kwargs_atom if kwargs_atom is not None else {}

    # create atom
    atom = Atom(parent._get_id(Atom), element, isotope, stereo, hydrogens, charge, valance, class_,
                parent=parent, **kwargs_atom)

    # create bonds
    new_bond1 = Bond(parent._get_id(Bond), bond_to_insert_into.symbol, bond_to_insert_into.atom1, atom, parent=parent)
    new_bond2 = Bond(parent._get_id(Bond), bond_to_insert_into.symbol, atom, bond_to_insert_into.atom2, parent=parent)

    # remove old bond and add new bonds and atoms to parent
    bond_index = parent.nodes.index(bond_to_insert_into)
    parent.nodes.insert(bond_index + 1, new_bond1)
    parent.nodes.insert(bond_index + 2, atom)
    parent.nodes.insert(bond_index + 3, new_bond2)
    parent.nodes.pop(bond_index)

    # remove old bond and add new bonds and atoms to root
    bond_index = parent.root.bonds.index(bond_to_insert_into)
    parent.root.bonds.insert(bond_index + 1, new_bond1)
    parent.root.bonds.insert(bond_index + 2, new_bond2)
    parent.root.bonds.pop(bond_index)
    atom_index = parent.root.atoms.index(bond_to_insert_into.atom1)
    parent.root.atoms.insert(atom_index + 1, atom)

    # remove old bond from atoms and add new bonds to atoms
    bond_to_insert_into.atom1.bonds.remove(bond_to_insert_into)
    bond_to_insert_into.atom2.bonds.remove(bond_to_insert_into)
    add_bond_to_connected_objects(new_bond1)
    add_bond_to_connected_objects(new_bond2)

    return parent


def add_bonding_descriptor_bond_via_index(
        parent: StochasticFragment,
        bond_symbol: str | None,
        descriptor: str,
        index_: int,
        prior_atom: Atom,
        kwargs_bond: dict = None,
):
    """

    Parameters
    ----------
    parent
    bond_symbol
    descriptor
    index_
    prior_atom
    kwargs_bond

    Returns
    -------

    """
    kwargs_bond = kwargs_bond if kwargs_bond is not None else {}

    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_, bond_symbol)
    bond = Bond(parent._get_id(Bond), bond_symbol, prior_atom, bd_atom, parent=parent, **kwargs_bond)
    add_bond_to_connected_objects(bond)

    if prior_atom == prior_atom.parent.nodes[0]:  # if first atom
        parent.nodes.insert(0, bond)
        parent.root.bonds.insert(0, bond)
        parent.nodes.insert(0, bd_atom)
        parent.root.atoms.insert(0, bd_atom)
    elif prior_atom == prior_atom.parent.nodes[-1] or \
            all([False for node in parent.nodes[parent.nodes.index(prior_atom) + 1:] if not isinstance(node, Branch)]):
        # if last atom or last atom followed by only branches
        parent.nodes.append(bond)
        parent.root.bonds.append(bond)
        parent.nodes.append(bd_atom)
        parent.root.atoms.append(bd_atom)
    else:
        branch = Branch(parent._get_id(Branch), parent)
        branch.nodes = [bond, bd_atom]
        insert_obj_into_list(parent.nodes, prior_atom, branch)
        insert_obj_into_list(parent.root.bonds, prior_atom.bonds[0], bond)
        insert_obj_into_list(parent.root.atoms, prior_atom, bd_atom)


def add_stochastic_fragment(stoch_object: StochasticObject, stoch_fragment: StochasticFragment) -> StochasticObject:
    """

    Parameters
    ----------
    stoch_object
    stoch_fragment

    Returns
    -------

    """
    stoch_object.nodes.append(stoch_fragment)
    stoch_fragment.parent = stoch_object

    for node in stoch_fragment.nodes:
        if isinstance(node, Atom):
            stoch_object.root.atoms.append(node)
        elif isinstance(node, Bond):
            stoch_object.root.bonds.append(node)
        elif isinstance(node, BondDescriptorAtom):
            # re-direct descriptors to the one in the stoch_object
            for descriptor in stoch_object.bonding_descriptors:
                if descriptor == node.descriptor:
                    node.descriptor = descriptor
                    continue

            # if descriptor not found already in stochastic object, add it
            stoch_object.bonding_descriptors.append(node.descriptor)

    return stoch_object


def add_bigsmiles_as_stochastic_fragment(stoch_obj: StochasticObject, bigsmiles_: BigSMILES):
    """

    Parameters
    ----------
    stoch_obj
    bigsmiles_

    Returns
    -------

    """
    stoch_fragment = StochasticFragment(stoch_obj.root._get_id(StochasticFragment), stoch_obj)
    stoch_fragment.nodes = bigsmiles_.nodes
    stoch_fragment.rings = bigsmiles_.rings
    stoch_fragment.bonding_descriptors = \
        list({node.descriptor for node in bigsmiles_.nodes if isinstance(node, BondDescriptorAtom)})

    return add_stochastic_fragment(stoch_obj, stoch_fragment)


def set_new_parent(new_parent, obj):
    """ re-direct 'parent' to new bigsmiles object """
    for node in obj.nodes:
        if hasattr(node, 'parent'):
            node.parent = new_parent


__all__ = ["has_node_attr", "add_atom", "add_bond", "add_ring_by_index", "add_ring_from_atoms", "add_bonding_descriptor_atom",
           "add_bond_atom_pair", "add_bond_bonding_descriptor_pair", "open_branch", "close_branch",
           "open_stochastic_object", "open_stochastic_object_fragment", "open_stochastic_object_fragment_with_bond",
           "close_stochastic_object", "open_stochastic_fragment", "close_open_stochastic_fragment",
           "close_stochastic_fragment", "exit_construction", "get_prior"
           ]



from __future__ import annotations
from bigsmiles.data_structures.bigsmiles import BigSMILES, Atom, Bond, Branch, StochasticFragment, \
    StochasticObject, BondDescriptorAtom


class Reaction:
    """

    data structure for reaction BigSMILES

    !!! info "Accepted patterns"

        * reactant '>' agent '>' product
        * reactant '>>' product
        * reactant.reactant>>product.product
        * reactant,reactant>>product,product (comma supported for easier parsing)

    !!! info "Definitions"
        **^^reactants:^^** materials that contributing one or more atoms to the product

        **^^agents:^^** materials that don't contribute any atoms to the product or receive atoms from the reactant
        (catalysts, solvents)

        **^^product:^^** output of reactions. all atoms should have come from the reactants

    """
    __slots__ = ["reactants", "agents", "products", "__dict__"]

    def __init__(self, text: str | None = None, **kwargs):
        """

        Parameters
        ----------
        text: str
            Reaction BigSMILES string
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.reactants: list[BigSMILES] = []
        self.agents: list[BigSMILES] = []
        self.products: list[BigSMILES] = []

        # parse input string
        if text:
            from bigsmiles.constructors.constructor_reaction import parse_reaction
            self.reactants, self.agents, self.products = parse_reaction(text)

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(Atom, Bond, Branch, StochasticFragment, StochasticObject, BondDescriptorAtom),
                              skip_color=True)

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = ""
        text += ",".join(chem.to_string(show_hydrogens, show_atom_index, show_repr, skip_color)
                         for chem in self.reactants)
        if self.agents:
            text += ">"
            text += ",".join(chem.to_string(show_hydrogens, show_atom_index, show_repr, skip_color)
                             for chem in self.agents)
            text += ">"
        else:
            text += ">>"
        text += ",".join(chem.to_string(show_hydrogens, show_atom_index, show_repr, skip_color)
                         for chem in self.products)

        return text



"""

Code for tokenizing a BigSMILES string.

"""
from __future__ import annotations
import enum
import re

from bigsmiles.errors import TokenizeError
import bigsmiles.reference_data.chemical_data as chemical_data


class TokenKind(enum.Enum):
    """
    **Enum**

    the kind of tokens that will be extracted from BigSMILES string

    Attributes
    ----------
    Bond
    Atom
    Aromatic
    AtomExtend
    BranchStart
    BranchEnd
    Ring
    Ring2
    BondEZ
    Disconnected
    Rxn
    BondDescriptor
    StochasticSeperator
    StochasticStart
    StochasticEnd
    ImplictEndGroup
    BondDescriptorLadder

    """
    Bond = 0
    Atom = 1
    Aromatic = 2
    AtomExtend = 3
    BranchStart = 4
    BranchEnd = 5
    Ring = 6
    Ring2 = 7
    BondEZ = 8
    Disconnected = 9
    Rxn = 10
    BondDescriptor = 11
    StochasticSeperator = 12
    StochasticStart = 13
    StochasticEnd = 14
    ImplictEndGroup = 15
    BondDescriptorLadder = 16


_isotope_pattern = r'(?P<isotope>[\d]{1,3})?'
_element_pattern = r'(?P<symbol>' + "|".join(chemical_data.aromatic_elements) + "|" \
                   + "|".join(chemical_data.elements_ordered) + ')'
_stereo_pattern = r'(?P<stereo>@{1,2})?'
_hydrogen_pattern = r'(?P<hydrogens>H[\d]?)?'
_charge_pattern = r'(?P<charge>[-|\+]{1,3}[\d]?)?'
_class_pattern = r'(?P<class_>:\d{1,3})?'
atom_pattern = r"(?:\[)" + _isotope_pattern + _element_pattern + _stereo_pattern + \
               _hydrogen_pattern + _charge_pattern + _class_pattern + r"(?:\])"

token_specification = [
    # order in the list is important; regex stops at first match
    (TokenKind.Bond.name, r'[-|=|#|$]'),
    (TokenKind.Atom.name, "|".join(chemical_data.organic_ordered)),
    (TokenKind.Aromatic.name, "|".join(chemical_data.aromatic_elements)),
    (TokenKind.AtomExtend.name, atom_pattern),  # Atom in brackets
    (TokenKind.BranchStart.name, r'\('),
    (TokenKind.BranchEnd.name, r'\)'),
    (TokenKind.Ring.name, r'[\d]{1}'),
    (TokenKind.Ring2.name, r'%[\d]{2}'),  # ring_id with two-digit numbers
    (TokenKind.BondEZ.name, r'/|\\'),  # cis trans
    (TokenKind.Disconnected.name, r"\."),  # mixture

    (TokenKind.BondDescriptorLadder.name, r"\[[$<>][\d]\[[$<>][\d]?\][\d]?\]"),  # Ladder
    (TokenKind.BondDescriptor.name, r"\[[$<>][\d]?[\d]?\]"),
    (TokenKind.StochasticSeperator.name, r",|;"),
    (TokenKind.StochasticStart.name, r'\{'),
    (TokenKind.StochasticEnd.name, r'\}'),
    (TokenKind.ImplictEndGroup.name, r'\[\]'),

    (TokenKind.Rxn.name, r">>|>"),  # reaction -->

    ('SKIP', r'[ \t]+'),  # Skip over spaces and tabs
    ('MISMATCH', r'.'),  # Any other character
]

tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_specification)

tok_regex = re.compile(tok_regex)


class Token:
    """
    token; what a BigSMILES string gets broken up into
    """
    __slots__ = ("kind", "value")

    def __init__(self, kind: TokenKind, value: str):
        """
        Parameters
        ----------
        kind: TokenKind
            token kind
        value: str
            token text

        Examples
        --------
        >>> Token(TokenKind.AtomExtend, "[13CH2]")

        """
        self.kind = kind
        self.value = value

    def __str__(self):
        return f"{self.kind.name}: {self.value}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.kind != other.kind:
            return False
        if self.value != other.value:
            return False
        return True


def tokenize(text: str) -> list[Token]:
    """

    tokenizes a bigSMILES string

    Parameters
    ----------
    text: str
        BigSMILES string

    Returns
    -------
    result: list[Token]
        BigSMILES as a token list

    Raises
    ------
    TokenizeError
        invalid symbol detected

    Examples
    --------
    >>> tokenize("C(C)C")
    [Token(TokenKind.Atom, "C"), Token(TokenKind.BranchStart, "("), Token(TokenKind.Atom, "C"),
    Token(TokenKind.BranchEnd, ")"), Token(TokenKind.Atom, "C")]

    """
    result = []
    for match in re.finditer(tok_regex, text.replace(" ", "")):
        kind = match.lastgroup

        value = match.group()
        if kind == 'SKIP':
            continue
        elif kind == 'MISMATCH':
            if text[:2] in chemical_data.element_symbols:
                raise TokenizeError(f"Invalid symbol. If the symbol is not in the list below it must be in []."
                                    f"\nNon-bracket elements: {chemical_data.organic_elements}"
                                    f"\n(starting with {value!r}; index: {match.span()[0]})"
                                f'\n{text}' + "\n" + " " * match.span()[0] + "^(and forward)")

            raise TokenizeError(f'Invalid symbol (or group of symbols). (starting with {value!r}; '
                                f'index: {match.span()[0]})'
                                f'\n{text}' + "\n" + " " * match.span()[0] + "^(and forward)")

        result.append(
            Token(TokenKind[kind], value)
        )

    return result


ATOM_PATTERN = re.compile(atom_pattern)


def tokenize_atom_symbol(symbol: str) -> dict:
    """
    convert atom symbol into a dictionary of the following values:

    * symbol
    * isotope (default: None)
    * stereo  (default: None)
    * hydrogens  (default: None)
    * charge  (default: 0)
    * class_  (default: None)

    Parameters
    ----------
    symbol: str
        atom string

    Returns
    -------
    result: dict[str: str | int]
        {"symbol": str, "isotope": int, "stereo": str, "hydrogens": int, "charge": int, "class_": int}

    Examples
    --------
    >>> tokenize_atom_symbol("[13C@H+:1]")
    {"symbol": "C", "isotope": 13, "stereo": "@", "hydrogens": 1, "charge": 1, "class_": 1}

    """
    if symbol in chemical_data.elements_aromatic:
        return {"symbol": symbol, "isotope": None, "stereo": None, "hydrogens": None, "charge": 0, "class_": None}

    try:
        results = ATOM_PATTERN.match(symbol).groupdict()

        if results["isotope"] is not None:
            results["isotope"] = int(results["isotope"])

        if results["hydrogens"] is None:
            results["hydrogens"] = 0
        else:
            if results["hydrogens"][-1].isdigit():
                results["hydrogens"] = int(results["hydrogens"][-1])
            else:
                results["hydrogens"] = 1

        if results["charge"] is None:
            results["charge"] = 0
        elif results["charge"] == "+":
            results["charge"] = 1
        elif results["charge"] == "-":
            results["charge"] = -1
        elif results["charge"].count("+") > 1:
            results["charge"] = results["charge"].count("+")
        elif results["charge"].count("-") > 1:
            results["charge"] = -1 * results["charge"].count("-")
        else:
            results["charge"] = int(results["charge"].replace("+", ""))

        if results["class_"] is not None:
            results["class_"] = int(results["class_"][1:])  # index 0 is ":", so skip it

        return results
    except AttributeError:
        raise TokenizeError(f"Issue tokenizing atom: {symbol}")


DEFAULT_BONDING_DESCRIPTOR_INDEX = 1


def tokenize_bonding_descriptor(symbol: str) -> tuple[str, int]:
    """
    convert bonding descriptor symbol into the following values:

    * symbol
    * index

    Parameters
    ----------
    symbol: str
        bonding_descriptor string

    Returns
    -------
    result: tuple[str, int]
        [symbol, index]

    Examples
    --------
    >>> tokenize_atom_symbol("[$1]")
    ["$", 1]

    Notes
    -----
    * default bonding descriptor index = 1

    """
    symbol = symbol.replace("[", "").replace("]", "")
    if not symbol:
        return symbol, DEFAULT_BONDING_DESCRIPTOR_INDEX

    if symbol[-1].isdigit():
        return symbol[0], int(symbol[1:])

    return symbol, DEFAULT_BONDING_DESCRIPTOR_INDEX



from __future__ import annotations
import re

import bigsmiles.errors as errors
from bigsmiles.data_structures.bigsmiles import BigSMILES, Bond, Atom


rxn_split_pattern = re.compile(r"(?<!\[)(>>|>)")


def parse_reaction(text: str) -> tuple[list[BigSMILES], list[BigSMILES], list[BigSMILES]]:
    """
    Main entry point to parsing reaction SMILES.

    !!! info "Regular Expression Pattern"

        pattern: r"(?<!\[)(>>|>)"

    Parameters
    ----------
    text: str
        reaction BigSMILES string

    Returns
    -------
    reactants: list[BigSMILES]
        reactants
    agents: list[BigSMILES]
        agents
    products: list[BigSMILES]
        products
    """
    result = re.split(rxn_split_pattern, text.replace(" ", ""))

    try:
        if ">>" in result:
            return parse_no_agent_reaction(result)
        return parse_agent_reaction(result)
    except errors.BigSMILESError as e:
        raise errors.BigSMILESError(f"Parsing failed on '{text}'.", e) from e


def parse_no_agent_reaction(text_list: list[str]) -> tuple[list[BigSMILES], list[BigSMILES], list[BigSMILES]]:
    """ the pattern should be [reactants, >>, products] """
    if len(text_list) != 3:
        raise errors.TokenizeError("Invalid BigSMILES reaction. Too many or few '>' or '>>' detected. "
                                   f"\nDetected chunks: {len(text_list)} (expected chunks: 3)")
    if ">" in text_list:
        raise errors.TokenizeError("Invalid BigSMILES reaction. Reactions must follow one of these patterns: "
                                   "\n\t 'reactants >> products'\n\t 'reactants > agents > products' ")
    if text_list[1] != ">>":
        raise errors.TokenizeError("Invalid BigSMILES reaction. More than one '>>' detected.")

    reactants = process_chemical_block(text_list[0])
    products = process_chemical_block(text_list[2])

    return reactants, [], products


def parse_agent_reaction(text_list: list[str]) -> tuple[list[BigSMILES], list[BigSMILES], list[BigSMILES]]:
    """ the pattern should be [reactants, >, agents, >, products] """
    if len(text_list) != 5:
        raise errors.TokenizeError("Invalid BigSMILES reaction. Too many or few '>' or '>>' detected. "
                                   f"\nDetected chunks: {len(text_list)} (expected chunks: 5)")
    if text_list[1] != ">" and text_list[3] != ">":
        raise errors.TokenizeError("Invalid BigSMILES reaction. "
                                   "The pattern should be: 'reactants > agents > products'.")

    reactants = process_chemical_block(text_list[0])
    agents = process_chemical_block(text_list[2])
    products = process_chemical_block(text_list[4])

    return reactants, agents, products


def process_chemical_block(text: str) -> list[BigSMILES]:
    text_list = comma_split(text)

    chemicals = []
    for chunk in text_list:
        chemical = BigSMILES(chunk)
        chemical = split_chemical(chemical)
        chemicals += chemical

    return chemicals


def comma_split(text: str) -> list[str]:
    """ splits on commas only at the top level (don't hit on commas in stochastic objects) """
    flag = 0
    buffer = ""
    result = []
    for char_ in text:
        if char_ == "{":
            flag += 1
        elif char_ == "}":
            flag -= 1
        elif char_ == "," and flag == 0:
            result.append(buffer)
            buffer = ""
            continue

        buffer += char_

    if buffer:
        result.append(buffer)

    return result


## Split chemicals with '.' disconnect notation ## noqa
#######################################################################################################################
def split_chemical(bigsmiles_: BigSMILES) -> list[BigSMILES]:
    """
    Tries to split a BigSMILES at disconnects '.' if it thinks they are different molecules/polymers.

    If it can't determine, it won't do anything.

    Parameters
    ----------
    bigsmiles_: BigSMILES
        bigsmiles which may or may not be made of multiple molecules/polymers seperated by disconnect notation '.'

    Returns
    -------
    results: list[BigSMILES]
        list of BigSMILES of individual molecules/polymers

    """
    if not bigsmiles_.has_disconnect:
        return [bigsmiles_]

    split_index = determine_split_index(bigsmiles_)
    if not split_index:
        return [bigsmiles_]

    return split_bigsmiles(bigsmiles_, split_index, True)


def determine_split_index(bigsmiles_: BigSMILES) -> list[int]:
    disconnect_index = []
    for i, node in enumerate(bigsmiles_.nodes):
        if isinstance(node, Bond) and node.bond_order == 0:
            disconnect_index.append(i)

    split_index = []  # list[int]
    buffer = 0
    can_split = True
    for index_ in disconnect_index:
        buffer_fragment = 0
        # check fragment
        for node in bigsmiles_.nodes[:index_]:
            if find_bridging_bond(node, bigsmiles_.nodes[index_].id_):
                can_split = False
                break

            if isinstance(node, Atom) and node.charge != 0:
                buffer_fragment += node.charge

        buffer += buffer_fragment
        if buffer == 0 and can_split:
            split_index.append(index_)

    return split_index


def find_bridging_bond(node, cut_off_index: int) -> bool:
    """
    check bonds for an atom that has an id_ greater than cut off index.
    cut off index is the disconnect symbol index so atom index is greater than the bond bridges across a disconnect
    """
    if not hasattr(node, "bonds"):
        return False

    for bond in node.bonds:
        if bond.bond_order == 0:
            continue
        for atom in bond:
            if atom.id_ > cut_off_index:
                return True  # found a bond that bridge across disconnect

    return False


def split_bigsmiles(bigsmiles_: BigSMILES, split_index: list[int], delete_index: bool = False) -> list[BigSMILES]:
    """ Brake bigsmiles up into multiple BigSMILES. """
    if not split_index:
        return [bigsmiles_]
    split_index.sort()

    result: list[BigSMILES] = []
    for index_ in reversed(split_index):
        new_bigsmiles = BigSMILES()

        # delete bond
        if delete_index:
            bigsmiles_.nodes[index_].delete()

        # move nodes
        new_bigsmiles.nodes += bigsmiles_.nodes[index_:]
        del bigsmiles_.nodes[index_:]

        # move atoms
        atom_index = get_index(new_bigsmiles.nodes, bigsmiles_.atoms, Atom)
        if atom_index is not None:
            new_bigsmiles.root.atoms += bigsmiles_.atoms[atom_index:]
            del bigsmiles_.atoms[atom_index:]

        # move bonds
        bond_index = get_index(new_bigsmiles.nodes, bigsmiles_.bonds, Bond)
        if bond_index is not None:
            new_bigsmiles.root.bonds += bigsmiles_.bonds[bond_index:]
            del bigsmiles_.bonds[bond_index:]

        # move rings
        ring_index = get_ring_index(new_bigsmiles, bigsmiles_)
        if ring_index is not None:
            new_bigsmiles.root.rings += bigsmiles_.rings[index_:]
            del bigsmiles_.rings[index_:]

        set_new_parent(new_bigsmiles, new_bigsmiles)
        re_number_node_ids(new_bigsmiles)
        result.append(new_bigsmiles)

    result.append(bigsmiles_)
    return list(reversed(result))


def set_new_parent(new_parent, obj):
    """ re-direct 'parent' to new bigsmiles object; only need to do the first layer """
    for node in obj.nodes:
        if hasattr(node, 'parent'):
            node.parent = new_parent


def get_index(node_list, other_list, type_) -> int | None:
    for node in node_list:
        if isinstance(node, type_):
            break
    else:
        return None

    for i, obj in enumerate(other_list):
        if obj == node:
            return i

    return None


def get_ring_index(new_bigsmiles: BigSMILES, old_bigsmiles: BigSMILES) -> int | None:
    for i, ring in enumerate(old_bigsmiles.rings):
        if ring.atom1 in new_bigsmiles.atoms:
            return i

    return None


def re_number_node_ids(obj, id_: int = 0):
    """ Recursive renumbering 'id_'. """
    root = obj.root
    for node in obj.nodes:
        node.id_ = root._get_id(type(node))
        if hasattr(node, 'nodes'):
            re_number_node_ids(node, id_)



"""

Remaining validation done on the BigSMILES object

Note:
* tokenization should take care of symbol validation
* Bigsmiles_constructor should take care of most bond and valence issues

This is mainly to catch additional leftover validation.

"""
from __future__ import annotations

from bigsmiles.data_structures.bigsmiles import BigSMILES, StochasticObject, Branch, BondDescriptor, Bond, has_node_attr
from bigsmiles.errors import ValidationError


def run_syntax_fixes(bigsmiles: BigSMILES):
    remove_unnecessary_branch_symbols(bigsmiles)
    check_ring_numbers(bigsmiles)


def remove_unnecessary_branch_symbols(obj: has_node_attr):
    """
    Checks if branch is notation is need; if not needed remove it
    Example: CCC(C(CC)) --> CCCCCC
    recursive function
    """
    for node in obj.nodes:
        # recursive call
        if hasattr(node, "nodes"):
            remove_unnecessary_branch_symbols(node)

    # do check
    if isinstance(obj.nodes[-1], Branch):
        # checks if branch is end of nodes; if it is then it's not needed
        branch = obj.nodes.pop()
        obj.nodes += branch.nodes


def check_ring_numbers(obj: has_node_attr):
    """ Renumber rings. """
    rings = _get_rings(obj, [])

    if not rings:
        return

    for i, ring in enumerate(rings):
        ring.ring_id = i + 1


def _get_rings(obj: has_node_attr, ring_list: list[Bond]) -> list[Bond]:
    """ Recursive get rings. """
    for node in obj.nodes:
        if hasattr(node, 'nodes'):
            _get_rings(node, ring_list)

    if hasattr(obj, 'rings'):
        ring_list += obj.rings
    return ring_list


def re_number_node_ids(obj: has_node_attr, id_: int = 0):
    """ Recursive renumbering 'id_'. """
    for node in obj.nodes:
        node.id_ = id_
        id_ += 1
        if hasattr(node, 'nodes'):
            re_number_node_ids(node, id_)


## Error checks ## noqa
#######################################################################################################################
def run_validation(bigsmiles: BigSMILES):
    """ Main entry point for post-construction validation. """
    check_ring_closure(bigsmiles)
    check_bonding_descriptors(bigsmiles)
    check_implicit_endgroups_ends(bigsmiles)


def check_ring_closure(bigsmiles: BigSMILES):
    """ Checks to see that all rings have been closed that were started. """
    for ring in bigsmiles.rings:
        if ring.atom2 is None:
            raise ValidationError(f"Ring opened, but not closed. (ring id: {ring.ring_id})")


def check_bonding_descriptors(bigsmiles: BigSMILES | StochasticObject | Branch):
    """

    rules:
    if [>] there must be [<]
    if [<] there must be [>]
    if end is implicit; there must be single bonding units
    if ends are explict they must both have a matching pair in the stochastic fragments



    Parameters
    ----------
    bigsmiles

    Returns
    -------

    """
    for node in bigsmiles.nodes:
        if isinstance(node, StochasticObject):
            _do_check_bonding_descriptors(node)

        # check nested stochastic objects
        if hasattr(node, "nodes"):
            check_bonding_descriptors(node)


def _do_check_bonding_descriptors(stoch_obj: StochasticObject):
    for bd in stoch_obj.bonding_descriptors:
        _ = bd.bond_symbol  # runs bond order check

        # if [$], [$0], [$1], ... must be 2 or more
        if bd.descriptor == "$":
            if len(bd.instances) < 1:
                raise ValidationError("[$] type bonding descriptors require more than one instances in a string.")

        # if [>] there must be [<] and if [<] there must be [>]
        if bd.descriptor in ("<", ">"):
            bd_pair = _find_complementing_bonding_descriptor(stoch_obj, bd)
            if bd_pair is None:
                raise ValidationError(f'{bd} complementary partner not found.')


def _find_complementing_bonding_descriptor(stoch_obj: StochasticObject, bond_descr: BondDescriptor) \
        -> BondDescriptor | None:
    for bd in stoch_obj.bonding_descriptors:
        if bond_descr.is_pair(bd):
            return bd

    return None


def check_implicit_endgroups_ends(obj: has_node_attr, parent_obj: has_node_attr = None):
    # if end is implicit; there must be single bonding units
    for i, node in enumerate(obj.nodes):
        if isinstance(node, StochasticObject):
            if node.bd_left.implicit:
                if i != 0:
                    # nothing allowed to the left
                    raise ValidationError("With a the left end group implicit, "
                                          "there should be nothing to the left of the stochastic object.")
                if parent_obj is not None:
                    # if it isn't BigSMILES it can't have a left implicit end group
                    raise ValidationError("Implicit left end group not allowed within interior.")

            if node.bd_right.implicit:

                if i != len(obj.nodes) - 1:
                    # nothing allowed to the right
                    raise ValidationError("With a implicit right end group, there should not be anything to the "
                                          "right of it.")

        if hasattr(node, "nodes"):
            check_implicit_endgroups_ends(node, obj)






def adjacency_list(bigsmiles_: BigSMILES):
    """
    a collection of unordered lists used to represent a finite graph
    Parameters
    ----------
    bigsmiles_

    Returns
    -------

    """
    adjacency_list = []






from __future__ import annotations

import logging

import bigsmiles.errors as errors
from bigsmiles.validation.validation_string import run_string_validation
from bigsmiles.constructors.tokenizer import Token, TokenKind, tokenize
from bigsmiles.validation.validation_tokens import run_token_validation
import bigsmiles.constructors.constructor_str as constructor
from bigsmiles.data_structures.bigsmiles import BigSMILES, Branch, StochasticFragment, Atom, StochasticObject, \
    has_node_attr, BondDescriptorAtom


def get_next_token(tokens: list[Token], text: str = "") -> Token:
    try:
        return tokens.pop(0)
    except IndexError:
        raise errors.BigSMILESError(f"{text} can't be at the end of a BigSMILES string.")


atom_tokens = {TokenKind.Atom, TokenKind.AtomExtend, TokenKind.Aromatic}
aromatic_acceptors = (Atom, StochasticObject, BondDescriptorAtom)
map_next_token_to_bond = {
    TokenKind.Aromatic: ":",
    TokenKind.Atom: "",
    TokenKind.AtomExtend: "",
    TokenKind.StochasticStart: "",
    TokenKind.Disconnected: ".",
    TokenKind.BondDescriptor: "",
}


## Functions for mapping tokens to BigSMILES ##
#######################################################################################################################
def map_atom(parent: has_node_attr, tokens: list[Token], token: Token):
    if isinstance(parent, BigSMILES) and not parent:
        return constructor.add_atom_str(parent, token.value)

    if isinstance(parent, StochasticFragment) and len(parent.nodes) == 0:
        return constructor.add_atom_str(parent, token.value)

    if token.kind is TokenKind.Aromatic and constructor.get_prior(parent, aromatic_acceptors).aromatic:
        return constructor.add_bond_atom_pair_str(parent, ":", token.value)

    return constructor.add_bond_atom_pair_str(parent, "", token.value)


def map_bond(parent: has_node_attr, tokens: list[Token], token: Token):
    next_token = get_next_token(tokens, "Bond")

    if next_token.kind in atom_tokens:
        return constructor.add_bond_atom_pair_str(parent, token.value, next_token.value)

    if next_token.kind is TokenKind.BondDescriptor:
        return constructor.add_bond_bonding_descriptor_pair_str(parent, token.value, next_token.value)

    if next_token.kind is TokenKind.StochasticStart:
        return map_stochastic_object_start(parent, tokens, next_token, token)

    if next_token.kind is TokenKind.Ring or next_token.kind is TokenKind.Ring2:
        return constructor.add_ring_by_index(parent, int(next_token.value.replace('%', '')), token.value)

    raise errors.BigSMILESError(f"Bond can't be followed by: {next_token.kind}")


def map_bond_descriptor(parent: has_node_attr, tokens: list[Token], token: Token):
    next_token = get_next_token(tokens, "Bond Descriptor")
    if next_token.kind is TokenKind.StochasticEnd:
        return map_stochastic_object_end(parent, tokens, token)

    if isinstance(parent, Branch) and next_token.kind != TokenKind.BranchEnd:
        # check to make sure the branch closes immediately
        raise errors.BigSMILESError("If Bonding Descriptors is the first Branch symbol, "
                                    "it can only be followed by Branch End.")

    if isinstance(parent, StochasticFragment) and len(parent.nodes) == 0:
        # first StochasticFragment symbol
        # get bond
        if next_token.kind is TokenKind.Bond:
            bond = next_token.value
        elif next_token.kind in map_next_token_to_bond:
            bond = map_next_token_to_bond[next_token.kind]
        elif next_token.kind in (TokenKind.Ring, TokenKind.Ring2):
            bond = None
        else:
            raise errors.BigSMILESError(f"Invalid Symbol following stochastic object close.")
        tokens.insert(0, next_token)

        return constructor.add_bonding_descriptor_atom_str(parent, token.value, bond)

    tokens.insert(0, next_token)
    if constructor.get_prior(parent, (Atom, StochasticObject, BondDescriptorAtom)).aromatic:
        bond = ":"
    else:
        bond = ""

    return constructor.add_bond_bonding_descriptor_pair_str(parent, bond, token.value)


def map_branch_start(parent: has_node_attr, tokens: list[Token], token: Token):
    if isinstance(parent, Branch) and len(parent.nodes) == 0:
        raise errors.BigSMILESError("Branch can't be the first object within a branch.")
    return constructor.open_branch(parent)


def map_branch_end(parent: has_node_attr, tokens: list[Token], token: Token):
    return constructor.close_branch(parent)


def map_ring(parent: has_node_attr, tokens: list[Token], token: Token):
    if constructor.get_prior(parent, (Atom, StochasticObject)).aromatic:
        bond = ":"
    else:
        bond = ""

    return constructor.add_ring_by_index(parent, int(token.value.replace('%', '')), bond)


def map_stochastic_object_start(parent: has_node_attr, tokens: list[Token], token: Token, bond_token: Token = None):
    next_token = get_next_token(tokens, "Stochastic objects start symbol")

    if next_token.kind is TokenKind.ImplictEndGroup:
        return constructor.open_stochastic_object_fragment_str(parent, next_token.value)

    if next_token.kind is not TokenKind.BondDescriptor:
        raise errors.BigSMILESError(f"Stochastic object starts must be followed an explict or implicit end group.")

    if isinstance(parent, BigSMILES) and not parent:
        parent = constructor.add_atom_str(parent, "[H]")

    if constructor.get_prior(parent, (Atom, StochasticObject)).aromatic:
        bond = ":"
    elif bond_token is None:
        bond = ""
    else:
        bond = bond_token.value

    return constructor.open_stochastic_object_fragment_with_bond_str(parent, bond, next_token.value)


def map_stochastic_object_end(parent: has_node_attr, tokens: list[Token], token: Token) -> has_node_attr:
    parent = constructor.close_stochastic_fragment_str(parent)

    # get bond
    if len(tokens) < 1:
        bond = None
    elif tokens[0].kind is TokenKind.Bond:
        bond = tokens[0].value
    elif tokens[0].kind in map_next_token_to_bond:
        bond = map_next_token_to_bond[tokens[0].kind]
    elif tokens[0].kind in (TokenKind.Ring, TokenKind.Ring2):
        bond = None
    else:
        raise errors.BigSMILESError(f"Invalid Symbol following stochastic object close.")

    return constructor.close_stochastic_object_str(parent, token.value, bond)


def map_bond_seperator(parent: has_node_attr, tokens: list[Token], token: Token) -> has_node_attr:
    return constructor.close_open_stochastic_fragment_str(parent)


def map_reaction(parent: has_node_attr, tokens: list[Token], token: Token):
    raise errors.BigSMILESError("Reaction Symbol detected. Use 'bigsmiles.Reaction()' to process this input.")


def map_disconnect(parent: has_node_attr, tokens: list[Token], token: Token):
    next_token = get_next_token(tokens, "Disconnect")
    if next_token.kind in atom_tokens:
        return constructor.add_bond_atom_pair_str(parent, ".", next_token.value)

    if next_token.kind is TokenKind.StochasticStart:
        return constructor.open_stochastic_object_fragment_with_bond_str(parent, ".", next_token.value)

    if next_token.kind is TokenKind.BondDescriptor:
        return constructor.add_bond_bonding_descriptor_pair_str(parent, ".", next_token.value)

    raise errors.BigSMILESError(f"Disconnect can't be followed by '{next_token.kind.name}'. ")


# def map_bondez():
#     # the pattern must be Atom() / Atom () = Atom () / Atom


def NotImplementedFunc(*args, **kwargs):
    raise NotImplementedError()


def SkipSymbol(parent: has_node_attr, tokens: list[Token], token: Token) -> has_node_attr:
    logging.warning(f"Symbol skipped: {token.value}")
    return parent


# this dict maps a token kind to a function that will process or map the symbol into a BigSMILES
map_token_functions = {
    TokenKind.Bond: map_bond,
    TokenKind.Atom: map_atom,
    TokenKind.Aromatic: map_atom,
    TokenKind.AtomExtend: map_atom,
    TokenKind.BranchStart: map_branch_start,
    TokenKind.BranchEnd: map_branch_end,
    TokenKind.Ring: map_ring,
    TokenKind.Ring2: map_ring,
    TokenKind.BondEZ: map_bond,
    TokenKind.Disconnected: map_disconnect,
    TokenKind.Rxn: map_reaction,
    TokenKind.BondDescriptor: map_bond_descriptor,
    TokenKind.StochasticSeperator: map_bond_seperator,
    TokenKind.StochasticStart: map_stochastic_object_start,
    TokenKind.StochasticEnd: map_stochastic_object_end,
    TokenKind.ImplictEndGroup: map_bond_descriptor,
    TokenKind.BondDescriptorLadder: NotImplementedFunc
}

valid_first_symbols = {TokenKind.Atom, TokenKind.AtomExtend, TokenKind.Aromatic, TokenKind.StochasticStart}


def tokens_to_bigsmiles(parent: has_node_attr, tokens: list[Token]):
    """
    Main loop for converting tokens into BigSMILES objects.

    Parameters
    ----------
    parent:
    tokens:
        BigSMILES tokens

    """
    num_tokens = len(tokens)

    # guard statements
    if num_tokens < 1:
        raise errors.BigSMILESError(f"No BigSMILES symbols detected.")
    if tokens[0].kind not in valid_first_symbols:
        raise errors.BigSMILESError(f"BigSMILES can't start with a '{tokens[0].kind.name}' symbol.")

    # main loop
    while tokens:
        token = tokens.pop(0)
        func = map_token_functions[token.kind]
        try:
            parent = func(parent, tokens, token)
        except errors.BigSMILESError as e:
            raise errors.BigSMILESError(f"Issue with token '{token}'. (token: {num_tokens - len(tokens) - 1})",
                                        e) from e


def parse_bigsmiles_str(input_text: str, bigsmiles: BigSMILES):
    """
    Main function that turns BigSMILES string tokens then into a BigSMILES object.

    Parameters
    ----------
    input_text:
        BigSMILES string
    bigsmiles: BigSMILES
        this BigSMILES will be built up

    """
    try:
        input_text = run_string_validation(input_text)
        tokens = tokenize(input_text)
        tokens = run_token_validation(tokens)
        tokens_to_bigsmiles(bigsmiles, tokens)
        constructor.exit_construction(bigsmiles)
    except errors.BigSMILESError as e:
        raise errors.BigSMILESError(f"Parsing failed on '{input_text}'.", e) from e



"""

Validation on SMILES or BigSMILES string.

"""
from __future__ import annotations
import re

from bigsmiles.errors import ValidationError


def run_string_validation(text: str) -> str:
    """ Main entry point for pre-construction validation. """
    branch_symbol_validation(text)
    stochastic_object_validation(text)
    brackets_validation(text)
    # text = validate_ring_numbering(text)  # Faster to do after tokenization
    return text


def branch_symbol_validation(text: str):
    """ Confirm that there is an equal number of '(' to ')'. """
    count = text.count('(') - text.count(')')
    if count != 0:
        if count > 0:
            raise ValidationError(f"Invalid BigSMILES. Missing {count} closing branch symbols ')'. "
                                  f"\n Invalid string: {text}")
        raise ValidationError(f"Invalid BigSMILES. Missing {count} opening branch symbols '('. "
                              f"\n Invalid string: {text}")


def stochastic_object_validation(text: str):
    """ Confirm that there is an equal number of '{' to '}'. """
    count = text.count('{') - text.count('}')
    if count != 0:
        if count > 0:
            raise ValidationError(f"Invalid BigSMILES. Missing {count} closing stochastic object symbols"
                                  + " '}'. " + f"\n Invalid string: {text}")
        raise ValidationError(f"Invalid BigSMILES. Missing {count} opening stochastic object symbols "
                              + "'{'. " + f"\n Invalid string: {text}")


def brackets_validation(text: str):
    """ Confirm that there is an equal number of '[' to ']'. """
    count = text.count('[') - text.count(']')
    if count != 0:
        if count > 0:
            raise ValidationError(f"Invalid BigSMILES. Missing {count} closing bracket symbols ']'. "
                                  f"\n Invalid string: {text}")
        raise ValidationError(f"Invalid BigSMILES. Missing {count} opening bracket symbols '['. "
                              f"\n Invalid string: {text}")


ring_digit_pattern = re.compile(r"(?<![H+><$-]|\[)(%\d{2}|\d)")
extract_bracket_pattern = re.compile(r"(\[)(.*?)(\])")

ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
len_ascii_lowercase = len(ascii_lowercase)


def get_next_letter(num: int) -> str:
    num_sym = int(num/len_ascii_lowercase)
    mod_sym = num % len_ascii_lowercase

    if num == 0:
        sym = ascii_lowercase[mod_sym]
    else:
        sym = ascii_lowercase[mod_sym] * (num_sym + 1)

    return "<" + sym + ">"


def validate_ring_numbering(text: str, renumber: bool = True):
    """
    Confirm that there is only two of a number.

    Note: Faster to do after tokenization
    """
    # No rings
    if text.count("1") == 0:
        return text

    # remove atoms with special stuff as it mess with ring parsing
    extract = dict()
    split_list = re.split(extract_bracket_pattern, text)
    for i in range(len(split_list)):
        if split_list[i] == "[":
            replacement_symbol = get_next_letter(i)
            extract[replacement_symbol] = split_list[i + 1]
            split_list[i + 1] = replacement_symbol

    text = "".join(split_list)

    ring_numbers = [int(num.replace("%", "")) for num in ring_digit_pattern.findall(text)]
    ring_number_set = set(ring_numbers)

    # check for more than two
    for num in list(ring_number_set):
        count = ring_numbers.count(num)
        if count != 2:
            if count == 1:
                raise ValidationError(f"Invalid BigSMILES. Missing ring index. Only one found for: '{num}'. "
                                      f"\n Invalid string: {text}")
            elif renumber and count % 2 == 0:
                # renumber duplicates
                text = _renumber_rings(text, num, max(ring_number_set) + 1)
                for i in range(1, int(count / 2) + 1):
                    ring_number_set.add(max(ring_number_set) + i)

            else:
                raise ValidationError(f"Invalid BigSMILES. More than 2 ring index found for: '{num}'. "
                                      f"\n Invalid string: {text}")

    # do replacement
    for k, v in extract.items():
        text = text.replace(k, v)
    return text


def _renumber_rings(text: str, num: int, new_num: int) -> str:
    if num > 9:
        pattern = r"(?<![H+><$-]|\[)" + f"(%{num})"
    else:
        pattern = r"(?<![H+><$-%]|\[)" + f"({num})"

    hit_index = [(m.start(0), m.end(0)) for m in re.finditer(pattern, text)]

    new_num += int(len(hit_index) / 2 - 2)
    first = True
    for hit in reversed(hit_index[2:]):  # skip the first two
        if new_num > 9:
            text = text[:hit[0]] + "%" + str(new_num) + text[hit[1]:]
        else:
            text = text[:hit[0]] + str(new_num) + text[hit[1]:]

        if first:
            first = False
        else:
            first = True
            new_num -= 1

    return text



from __future__ import annotations
import logging

from bigsmiles.errors import ValidationError
from bigsmiles.constructors.tokenizer import Token, TokenKind


def run_token_validation(tokens: list[Token]) -> list[Token]:
    tokens = validate_ring_numbering(tokens)

    return tokens


def validate_ring_numbering(tokens: list[Token], renumber: bool = True) -> list[Token]:
    """ Confirm that there is only two of a number. """

    # get ring_numbers
    ring_numbers = dict()  # key: ring number; value: list[index in tokens]
    for i, token in enumerate(tokens):
        if token.kind is TokenKind.Ring:
            value = int(token.value)
        elif token.kind is TokenKind.Ring2:
            value = int(token.value.replace("%", ""))
        else:
            continue

        if value in ring_numbers:
            ring_numbers[value].append(i)
        else:
            ring_numbers[value] = [i]

    # No rings
    if not ring_numbers:
        return tokens

    # check for more than two
    max_index = max([k for k in ring_numbers.keys()]) + 1
    for key in list(ring_numbers.keys()):
        count = len(ring_numbers[key])
        if count == 2:
            continue
        elif count == 1:
            raise ValidationError(f"Invalid BigSMILES. Missing ring index. Only one found for: '{key}'. ")
        elif renumber and count % 2 == 0:
            logging.warning(f"Duplicate ring index detected ({key}) and fixed for: "
                            f"{''.join([str(t.value) for t in tokens])}")
            # renumber duplicates
            for i, index_ in enumerate(ring_numbers[key]):
                if int(i/2) == 0:
                    continue  # don't change the first set

                tokens[index_].value = str(max_index) if max_index < 10 else "%" + str(max_index)
                if i % 2 != 0:
                    max_index += 1
        else:
            raise ValidationError(f"Invalid BigSMILES. More than 2 ring index found for: '{key}'. ")

    return tokens



from __future__ import annotations


class BigSMILESError(Exception):
    """ base BigSMILES error """

    def __init__(self, text: str, error=None):
        if isinstance(error, BigSMILESError):
            text += '\n\t'
            text += error.text.replace("\t", "\t\t")
        self.text = text

    def __str__(self):
        return self.text


class TokenizeError(BigSMILESError):
    """ raised when an error occurs tokenizing a BigSMILES string. """


class ConstructorError(BigSMILESError):
    """ raised when an error occurs during the construction of a BigSMILES string. """


class ValidationError(BigSMILESError):
    """ raised when validating a SMILES/BigSMILES string is syntactically correct. """


ERRORS = (BigSMILESError, TokenizeError)



"""

Code for determining double bond E/Z stereo-chemistry

"""

from __future__ import annotations

import enum
import logging

import bigsmiles.reference_data.chemical_data as chemical_data
from bigsmiles.data_structures.bigsmiles import Bond, Atom, StochasticObject, BondDescriptorAtom


MAX_DEPTH = 100  # 100 is the max depth of traversal
DEFAULT_SCORE = 150


def get_other_node(bond: Bond, not_this_node: Atom | StochasticObject) \
        -> Atom | StochasticObject | BondDescriptorAtom | None:
    for atom in bond:
        if atom is not not_this_node:
            return atom

    return None


def get_other_bonds(atom: Atom, not_this_bond: tuple[Bond, ...] | list[Bond, ...]) -> list[Bond, ...]:
    bonds = []
    for bond in atom.bonds:
        if bond not in not_this_bond:
            bonds.append(bond)

    return bonds


def get_bond_with_ez_symbol(atom: Atom) -> Bond | None:
    ez_bond = None
    for bond in atom.bonds:
        if bond.symbol == "/" or bond.symbol == "\\":
            if ez_bond is not None:
                logging.warning(r"Extra '\' or '\\' detected." +
                                f"\nIgnoring symbol: {ez_bond.details} \nBigSMILES: {atom.root}")

            ez_bond = bond

    return ez_bond


class PriorityScore:
    __slots__ = ["atomic_number", "next_atom_highest_atomic_number", "number_bonds"]

    def __init__(self, atomic_number: int, next_atom_highest_atomic_number: int, number_bonds: float | int):
        self.atomic_number = atomic_number
        self.next_atom_highest_atomic_number = next_atom_highest_atomic_number
        self.number_bonds = number_bonds

    def __str__(self):
        return ",".join((str(getattr(self, name)) for name in self.__slots__))

    def __eq__(self, other):
        if self.atomic_number != other.atomic_number or self.number_bonds != other.number_bonds:
            return False

        return True

    def __lt__(self, other):
        if self.atomic_number < other.atomic_number:
            return True

        if self.atomic_number == other.atomic_number:
            if self.next_atom_highest_atomic_number < other.next_atom_highest_atomic_number:
                return True

            if self.next_atom_highest_atomic_number == other.next_atom_highest_atomic_number:
                if self.number_bonds < other.number_bonds:
                    return True

        return False


bond_scores = {
    ".": 0,
    "": 1,
    "-": 1,
    "/": 1,
    "\\": 1,
    ":": 1.5,
    "=": 2.1,  # little extra is to give them priority over two single bonds
    "#": 3.2,  # little extra is to give them priority over two single bonds
    "$": 4.1
}


def get_atom_score(node: Atom | StochasticObject | BondDescriptorAtom) -> int:
    if isinstance(node, Atom) and node.symbol in chemical_data.symbol_to_atomic_number:
        return chemical_data.symbol_to_atomic_number[node.symbol]

    return DEFAULT_SCORE


def get_score(node: Atom | StochasticObject | BondDescriptorAtom | None) -> PriorityScore:
    if node is None:
        return PriorityScore(0, 0, 0)

    # atom score
    atomic_number = get_atom_score(node)

    # next atom score
    next_atoms = (get_other_node(bond, node) for bond in node.bonds)
    next_atom_highest_atomic_number = max(get_atom_score(atom) for atom in next_atoms)

    # bond score
    if isinstance(node, Atom):
        number_bonds = sum(bond_scores[bond.symbol] for bond in node.bonds)
    elif isinstance(node, StochasticObject):
        number_bonds = node.bond_left.bond_order
    elif isinstance(node, BondDescriptorAtom):
        number_bonds = 0
    else:
        raise NotImplementedError("code bug")

    return PriorityScore(atomic_number, next_atom_highest_atomic_number, number_bonds)


class TraversalPath:

    def __init__(self, nodes: list = None, edges: list = None):
        self.nodes = nodes if nodes is not None else []
        self.edges = edges if edges is not None else []

    def __str__(self):
        return "->".join(repr(node) for node in self.nodes)

    def get_score(self) -> PriorityScore:
        return get_score(self.nodes[-1])


def next_step(path:  TraversalPath) -> list:
    additional_paths = []
    current_node = path.nodes[-1]

    if current_node is None:
        path.nodes.append(None)
        return additional_paths

    elif isinstance(current_node, Atom):
        atoms = []
        bonds = []
        for bond in current_node.bonds:
            if bond in path.edges:
                continue
            atoms.append(get_other_node(bond, current_node))
            bonds.append(bond)

        if len(atoms) == 0:
            path.nodes.append(None)
            return additional_paths

        atoms, bonds = sort_atoms_by_score(atoms, bonds)
        for i in range(len(atoms)):
            if i == 0:
                path.nodes.append(atoms[0])
                path.edges.append(bonds[0])
            else:
                additional_paths.append(
                    TraversalPath(path.nodes[:-1] + [atoms[i]], path.edges[:-1] + [bonds[i]])
                )

    elif isinstance(current_node, BondDescriptorAtom):
        path.nodes.append(None)

    elif isinstance(current_node, StochasticObject):
        if current_node.bond_left in path.edges:
            if current_node.bond_right is None:
                path.nodes.append(None)

            path.nodes.append(get_other_node(current_node.bond_right, current_node))
            path.edges.append(current_node.bond_right)

        else:
            if current_node.bond_left is None:
                path.nodes.append(None)

            path.nodes.append(get_other_node(current_node.bond_left, current_node))
            path.edges.append(current_node.bond_left)

    else:
        raise NotImplementedError("coding bug")

    return additional_paths


def sort_atoms_by_score(atoms: list[Atom, ...], bonds: list[Bond, ...]) -> tuple[list[Atom, ...], list[Bond, ...]]:
    score = []
    for atom in atoms:
        score.append(get_atom_score(atom))

    return [x for _, x in sorted(zip(score, atoms))], [x for _, x in sorted(zip(score, bonds))],


class TraversalDirection:

    def __init__(self, start_atom: Atom, bond_prior: Bond):
        self.start_atom = start_atom
        self.bond_prior = bond_prior
        self.paths: [TraversalPath] = [TraversalPath([start_atom], [bond_prior])]
        self.best_path = self.paths[0]
        self._scores = None
        self._up_to_date = False

    @property
    def num_paths(self) -> int:
        return len(self.paths)

    def scores(self) -> list[PriorityScore]:
        if not self._up_to_date:
            self._scores = [path.get_score() for path in self.paths]
            index_not_none = next(i for i, score in enumerate(self._scores) if score is not None)
            self.best_path = self.paths[index_not_none]
            self._up_to_date = True

        return self._scores

    def one_traverse_step(self):
        self._up_to_date = False

        # traverse the graph one step
        for i in range(len(self.paths)):  # index used as size of self.up may change during loop
            additional_paths = next_step(self.paths[i])
            self.paths += additional_paths


class TraversalOutcomes(enum.Enum):
    up = 1
    down = 2
    tie = 3


class TraversalSide:
    def __init__(self, double_bond: Bond, side: bool):
        """
        Parameters
        ----------
        double_bond:

        side:
            True: left, False: right
        """
        self.double_bond = double_bond
        self.side = side

        self.up: TraversalDirection | None = None
        self.down: TraversalDirection | None = None
        self.winner: TraversalOutcomes | None = None
        self.number_of_double_bonds_in_series: int = 0  # in a row after the starting double bond
        self.setup_successful: bool = True

        self.setup()

    def setup(self):
        if self.side:
            atom = self.double_bond.atom1
        else:
            atom = self.double_bond.atom2

        # get starting atom
        starting_atom = self.check_for_multiple_double_bonds(atom, self.double_bond)
        if starting_atom is None:
            return

        # determine up down directions
        if self.side:
            self.determine_up_down_atom_left(starting_atom)
        else:
            self.determine_up_down_atom_right(starting_atom)

    def check_for_multiple_double_bonds(self, atom: Atom, bond: Bond) -> Atom | None:
        """
        check for multiple double bonds in a row
        Maybe recursive

        Parameters
        ----------
        atom:
            current atom position
        bond:
            prior bond

        Returns
        -------
        atom:
            Atom to start up/down traversals from

        """
        other_bonds = get_other_bonds(atom, [bond])
        if not other_bonds:
            self.setup_successful = False
            return None  # double bond is terminated with implicit hydrogens or lone pairs
        if len(other_bonds) == 1 and other_bonds[0].symbol == "=":
            next_atom = get_other_node(other_bonds[0], atom)
            self.number_of_double_bonds_in_series += 1
            return self.check_for_multiple_double_bonds(next_atom, other_bonds[0])

        return atom

    def determine_up_down_atom_left(self, start_atom: Atom):
        bond_ez = get_bond_with_ez_symbol(start_atom)
        if bond_ez is None:
            self.setup_successful = False
            return  # no stereo bonds '/' or '\\'

        atom_ez = get_other_node(bond_ez, start_atom)

        # set up path
        direction = TraversalDirection(atom_ez, bond_ez)
        if bond_ez.symbol == "/":
            if atom_ez.parent is start_atom.parent:
                self.up = direction  # main chain
            else:
                self.down = direction  # branch
        else:  # "\\"
            if atom_ez.parent is start_atom.parent:
                self.down = direction  # main chain
            else:
                self.up = direction  # branch

        self.figure_out_other_bond(start_atom, bond_ez)

    def determine_up_down_atom_right(self, start_atom: Atom):
        bond_ez = get_bond_with_ez_symbol(start_atom)
        if bond_ez is None:
            self.setup_successful = False
            return  # no stereo bonds '/' or '\\'

        atom_ez = get_other_node(bond_ez, start_atom)

        # set first atom that has stereo symbol
        if bond_ez.symbol == "/":
            self.down = TraversalDirection(atom_ez, bond_ez)
        else:  # "\\"
            self.up = TraversalDirection(atom_ez, bond_ez)

        self.figure_out_other_bond(start_atom, bond_ez)

    def figure_out_other_bond(self, start_atom: Atom, bond_ez: Bond):
        # figure out other bond
        other_bond = get_other_bonds(start_atom, (self.double_bond, bond_ez))
        if len(other_bond) == 1:
            other_atom = get_other_node(other_bond[0], start_atom)
            if self.down is None:
                self.down = TraversalDirection(other_atom, other_bond[0])
            else:
                self.up = TraversalDirection(other_atom, other_bond[0])

        elif len(other_bond) == 0:
            # other bond must be implicit hydrogens or terminal
            if self.down is None:
                self.winner = TraversalOutcomes.up
            else:
                self.winner = TraversalOutcomes.down

        else:
            raise NotImplementedError("more than 2 bonds coming off double bond atom")

    def check_for_winner(self):
        if self.winner:
            return

        scores_up = self.up.scores()
        scores_down = self.down.scores()
        for i in range(max([len(scores_up), len(scores_down)])):
            if i > len(scores_up) - 1 and i > len(scores_up) - 1:
                self.winner = TraversalOutcomes.tie
            if i > len(scores_up) - 1:
                self.winner = TraversalOutcomes.down
                break
            if i > len(scores_down) - 1:
                self.winner = TraversalOutcomes.up
                break

            if scores_up[i] == scores_down[i]:
                if scores_up is None:
                    continue
                else:
                    break

            if scores_up > scores_down:
                self.winner = TraversalOutcomes.up
            else:
                self.winner = TraversalOutcomes.down

    def run_graph_traversal(self):
        self.check_for_winner()
        if self.winner:
            return

        logging.debug(f" Starting '{'left' if self.side else 'right'}' E/Z graph traversal:"
                      f" {self.double_bond.root.to_string(show_repr=(Atom,))}"
                      f"\n-------------------------------")
        logging.debug(f"\t {'step':>4}, {'up':<30}, {'num_up_paths':>12}, {'down':>30}, {'num_down_paths':>12}")
        logging.debug(f"\t {0:4}, {str(self.up.best_path):<30}, {self.up.num_paths:>12}"
                      f" {str(self.down.best_path):<30}, {self.down.num_paths:>12}")

        # perform graph traversal
        for i in range(1, MAX_DEPTH):
            self.up.one_traverse_step()
            self.down.one_traverse_step()

            self.check_for_winner()
            logging.debug(f"\t {i:4}, {str(self.up.best_path):<30}, {self.up.num_paths:>12}"
                          f" {str(self.down.best_path):<30}, {self.down.num_paths:>12}")

            if self.winner:
                return

        else:
            logging.warning(f'Max traversal depth reached on E/Z determination of: '
                            f'\nBigSMILES: {self.double_bond.parent} '
                            f'\nBond: {repr(self.double_bond)}')
            return


def check_setup(left_side: TraversalSide, right_side: TraversalSide) -> bool:
    """ Check to see if there were reasons to terminate early from traversal setup """
    if not left_side.setup_successful or not right_side.setup_successful:
        return True

    # check for 'odd' number of double bonds in a row
    if (left_side.number_of_double_bonds_in_series + right_side.number_of_double_bonds_in_series + 1) % 2 == 0:
        # left + right + 1 (starting double bond) = even  --> return None for E/Z
        logging.warning("Double bond stereo-chemistry notation '/' and '\\' should "
                        f"not be used on even number of double bonds; they are axial chiral and use @/@@ notation on "
                        f"center atom.\nBigSMILES:{left_side.double_bond.parent}")
        return True

    return False


def determine_winner(left_side: TraversalSide, right_side: TraversalSide) -> str | None:
    if left_side.winner is None or right_side.winner is None:
        return None

    if left_side.winner is TraversalOutcomes.tie and right_side.winner is TraversalOutcomes.tie:
        logging.warning("Double bond stereo-chemistry notation '/' and '\\' should "
                        "not be used on symmetric molecules.")
        return None

    if (left_side.winner is TraversalOutcomes.tie or right_side.winner is TraversalOutcomes.tie) or \
            (left_side.winner == right_side.winner):
        return "Z"

    return "E"


def get_double_bond_ez(double_bond: Bond) -> str | None:
    """
    Determines the E/Z stereochemistry of a double Bond

    Parameters
    ----------
    double_bond:
        double bond to start analysis at

    Returns
    -------
    result:
        will return 'E'/'Z' for odd number of double bonds, None for even number of double bonds (axial chiral)

    Notes
    -----

    * Breath-first-search (BFS) prioritizing heavy atoms and following more bonds.
    Starts out at the double bond with two traverses heading left and two right,
    with one going up and the other going down for a total of 4 simultaneous BFS.
    At each step the below rules are applied to compute a score
    The searches terminate once a difference between the up-down score.
    * Rules:
        * rule 1: higher atomic mass has higher priority
        * rule 2: (if same atomic mass) atom with more of these substituent
        * rule 3: propagate down chain till there is a difference
        * double/triple bonds are considered to be bonded to an equivalent single bonds C=C --> C(C)-C(C)
    * stochastic objects are scored as 150 and treated as a single atom
    * max traversal depth is 200

    """
    left_side = TraversalSide(double_bond, True)
    right_side = TraversalSide(double_bond, False)

    if check_setup(left_side, right_side):
        return None

    left_side.run_graph_traversal()
    right_side.run_graph_traversal()

    return determine_winner(left_side, right_side)



"""

Extend existing construction functions for string input

"""
from __future__ import annotations
from functools import wraps

from bigsmiles.errors import BigSMILESError
from bigsmiles.constructors.tokenizer import tokenize_atom_symbol, tokenize_bonding_descriptor
from bigsmiles.data_structures.bigsmiles import StochasticObject, StochasticFragment, BigSMILES, Branch
from bigsmiles.constructors.constructor import *


def in_stochastic_object(func):
    """ Decorator to ensure function call only occurs within stochastic object."""

    @wraps(func)
    def _in_stochastic_object(*args, **kwargs):
        if not args[0].in_stochastic_object:
            raise BigSMILESError("Must in stochastic object.")
        return func(*args, **kwargs)

    return _in_stochastic_object


## Extend existing construction functions for string input ##
#######################################################################################################################
def add_atom_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment, symbol: str) \
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
    appends an `Atom` to the end of the `parent`

    Parameters
    ----------
    parent:
        parent Atom will be added to
    symbol:
        atom symbol - will be parsed into symbol, isotope, charge, etc.

    Returns
    -------
    parent:
        returns same parent that was provided

    """
    return add_atom(parent, **tokenize_atom_symbol(symbol))


@in_stochastic_object
def add_bonding_descriptor_atom_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                                    bd_symbol: str, bond_symbol: str | None = None)\
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
    appends an `BondDescriptorAtom` to the end of the `parent`

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
        parent that the BondDescriptorAtom will be added to
    bd_symbol: str
        bonding descriptor symbol - will be parsed into symbol, index
    bond_symbol: str
        bond symbol ("", "=", "#", etc.)

    Returns
    -------
    parent:
        returns same parent that was provided

    """
    return add_bonding_descriptor_atom(parent, *tokenize_bonding_descriptor(bd_symbol), bond_symbol=bond_symbol)


def add_bond_atom_pair_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                           bond_symbol: str | None, atom_symbol: str) \
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
     appends an `Bond` followed by an 'Atom' to the end of the `parent`

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
        parent that the Bond and Atom will be added to
    bond_symbol: str
        bond symbol ("", "=", "#", etc.)
    atom_symbol: str
        atom symbol - will be parsed into symbol, isotope, charge, etc.

    Returns
    -------
    parent:
        returns same parent that was provided

    """
    return add_bond_atom_pair(parent, bond_symbol, **tokenize_atom_symbol(atom_symbol))


@in_stochastic_object
def add_bond_bonding_descriptor_pair_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                                         bond_symbol: str | None, bd_symbol: str)\
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
    appends an `Bond` followed by an 'BondDescriptorAtom' to the end of the `parent`

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
        parent that the Bond and BondDescriptorAtom will be added to
    bond_symbol: str
        bond symbol ("", "=", "#", etc.)
    bd_symbol: str
        bonding descriptor symbol - will be parsed into symbol, index

    Returns
    -------
    parent:
        returns same parent that was provided

    """
    return add_bond_bonding_descriptor_pair(parent, bond_symbol, *tokenize_bonding_descriptor(bd_symbol))


def open_stochastic_object_fragment_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                                        bd_symbol: str) -> StochasticFragment:
    """
    opens 'StochasticObject' followed by opening 'StochasticFragment'

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
    bd_symbol: str
        bonding descriptor symbol - will be parsed into symbol, index

    Returns
    -------
    parent: StochasticFragment
        returns parent is stochastic fragment

    """
    return open_stochastic_object_fragment(parent, *tokenize_bonding_descriptor(bd_symbol))


def open_stochastic_object_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                               bd_symbol: str) -> StochasticObject:
    """
    opens 'StochasticObject'

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
    bd_symbol: str
        bonding descriptor symbol - will be parsed into symbol, index

    Returns
    -------
    parent: StochasticFragment
        returns parent is stochastic object

    """
    return open_stochastic_object(parent, *tokenize_bonding_descriptor(bd_symbol))


def open_stochastic_object_fragment_with_bond_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment,
                                                  bond_symbol: str | None, bd_symbol: str) \
        -> StochasticFragment:
    """
    opens 'StochasticObject' followed by opening 'StochasticFragment'

    Parameters
    ----------
    parent: BigSMILES, Branch, StochasticFragment
        parent of StochasticObject
    bd_symbol: str
        bonding descriptor symbol - will be parsed into symbol, index
    bond_symbol: str
        bond symbol ("", "=", "#", etc.)

    Returns
    -------
    parent: StochasticFragment
        returns parent is StochasticFragment

    """
    return open_stochastic_object_fragment_with_bond(parent, bond_symbol, *tokenize_bonding_descriptor(bd_symbol))


def close_stochastic_object_str(parent: BigSMILES | Branch | StochasticObject | StochasticFragment, bd_symbol: str,
                                bond_symbol: str | None) \
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
    closes 'StochasticObject' and append it to `parent`

    Parameters
    ----------
    parent: StochasticObject
        parent of StochasticObject
    bd_symbol: str
        bonding descriptor symbol (right side) - will be parsed into symbol, index
    bond_symbol: str
        bond symbol ("", "=", "#", etc.)

    Returns
    -------
    parent.parent: BigSMILES, Branch, StochasticFragment
        returns parent of StochasticObject
    """
    return close_stochastic_object(parent, *tokenize_bonding_descriptor(bd_symbol), bond_symbol=bond_symbol)


@in_stochastic_object
def open_stochastic_fragment(parent) -> StochasticFragment:
    """
    opens 'StochasticFragment'

    Parameters
    ----------
    parent: StochasticObject
        parent of StochasticFragment

    Returns
    -------
    parent: StochasticFragment
        returns the freshly opened StochasticFragment

    """
    return open_stochastic_fragment(parent)


@in_stochastic_object
def close_open_stochastic_fragment_str(parent) -> StochasticFragment:
    """
    closes `StochasticFragment` and appends it to 'parent.parent' and opens a new `StochasticFragment`

    Parameters
    ----------
    parent: StochasticFragment
        StochasticFragment to be closed

    Returns
    -------
    parent: StochasticFragment
        returns the freshly opened StochasticFragment

    """
    return close_open_stochastic_fragment(parent)


@in_stochastic_object
def close_stochastic_fragment_str(parent: StochasticFragment) \
        -> BigSMILES | Branch | StochasticObject | StochasticFragment:
    """
    closes `StochasticFragment` and appends it to 'parent.parent'

    Parameters
    ----------
    parent: StochasticFragment
        StochasticFragment to be closed

    Returns
    -------
    parent.parent: StochasticObject
        returns the parent.parent

    """
    return close_stochastic_fragment(parent)



from __future__ import annotations
import os
import json


bond_mapping = {
    None: None,
    ".": 0,
    "": 1,
    "-": 1,
    "/": 1,
    "\\": 1,
    ":": 1.5,
    "=": 2,
    "#": 3,
    "$": 4
}

stereo_bonds = {"/", "\\"}

# load symbol data from JSON
file_path = os.path.dirname(os.path.realpath(__file__))
path_to_data = os.path.join(file_path, "elements_data.json")
with open(path_to_data, "r", encoding="UTF-8") as f:
    element_data = json.load(f)


class Element:
    __slots__ = ("name", "atomic_number", "atomic_mass", "period", "symbol", "electron_configuration", "group",
                 "aromatic", "organics", "conflict", "valences")

    def __init__(self,
                 name: str,
                 atomic_number: int,
                 atomic_mass: float,
                 period: int,
                 symbol: str,
                 electron_configuration: str,
                 group: str,
                 aromatic: bool,
                 organics: bool,
                 conflict: bool,
                 valences: tuple[int]
                 ):
        self.name = name
        self.atomic_number = atomic_number
        self.atomic_mass = atomic_mass
        self.period = period
        self.symbol = symbol
        self.electron_configuration = electron_configuration
        self.group = group
        self.aromatic = aromatic
        self.organics = organics
        self.conflict = conflict
        self.valences = valences

    def __str__(self):
        return self.symbol

    def __repr__(self):
        return self.symbol


periodic_table = tuple(Element(atomic_number=int(k), **v) for k, v in element_data.items())
del element_data

# create tuples, sets, and dictionaries for performance
element_symbols = {element.symbol for element in periodic_table}
elements_ordered = sorted(element_symbols, reverse=True)  # sorted in reverse so "Cl" hits before "C"
aromatic_elements = {element.symbol.lower() for element in periodic_table if element.aromatic}
elements_aromatic = element_symbols.union(aromatic_elements)
organic_elements = {element.symbol for element in periodic_table if element.organics}
organic_ordered = sorted(organic_elements, reverse=True)  # sorted in reverse so "Cl" hits before "C"
atom_valences = {element.symbol: element.valences for element in periodic_table}
symbol_to_atomic_number = {element.symbol: element.atomic_number for element in periodic_table}



from __future__ import annotations
import re

class MolecularFormula:
    def __init__(self, formula: str):
        self._formula = None
        self._elements = None

        self.formula = formula

    def __str__(self):
        return str(self.formula)

    def __repr__(self):
        return str(self.formula)

    def __getitem__(self, key: str):
        if key in self._elements:
            return self._elements[key]

    def __iter__(self):
        for element, num in self._elements.items():
            yield element, num

    @property
    def formula(self):
        return self._formula

    @formula.setter
    def formula(self, formula: str):
        if not isinstance(formula, str):
            raise TypeError(f"'formula' must be a str. ({type(formula)} given)")

        self._formula, self._elements = self.reduce(formula)

    @property
    def elements(self):
        return self._elements

    def reduce(self, formula: str) -> (str, dict[str, int]):
        """
        Reduces all chemical formulas to the same reduced version.
        :param formula:
        :return:
        """
        split_formula = self._split_formula(formula)
        elements = self._split_formula_to_element_dict(split_formula)
        formula = self._element_to_formula(elements, config_chemical.molecular_formula_element_order)

        return formula, elements

    @staticmethod
    def _split_formula(formula: str) -> list[str]:
        """
        Split formula up into letters and numbers.
        Example: "C1H3OH63Cr2CCCOOO" -> ['C', '1', 'H', '3', 'O', 'H', '63', 'Cr', '2', 'C', 'C', 'C', 'O', 'O', 'O']
        """
        pattern = r'[A-Z]{1}[a-z]*|\d+'
        return re.findall(pattern, formula)

    @staticmethod
    def _split_formula_to_element_dict(split_formula: list[str]) -> dict[str, int]:
        """
        Given a split chemical formula return a dictionary of element_symbols.
        :param split_formula:
        :return: {"symbol symbol": "count"} -> {"C": 1, "O": 2}
        """
        # put data into dictionary
        elements = {}
        for index, entry in enumerate(split_formula):
            if not entry.isnumeric():
                if entry not in periodic_table.symbols:
                    raise MolecularFormulaError(f"Invalid symbol in chemical formula. Invalid: {entry}")

                if entry in elements:
                    if index + 1 < len(split_formula):
                        num_ = split_formula[index + 1]
                    else:
                        num_ = "1"
                    if num_.isnumeric():
                        elements[entry] = int(num_) + elements[entry]
                    else:
                        elements[entry] = 1 + elements[entry]
                else:
                    if index + 1 < len(split_formula):
                        num_ = split_formula[index + 1]
                    else:
                        num_ = "1"
                    if num_.isnumeric():
                        elements[entry] = int(num_)
                    else:
                        elements[entry] = 1

        return elements

    def _element_to_formula(self, elements: dict, order: list[str] = None) -> str:
        """
        Converts symbol dictionary into a reduced string.
        """
        # Set order
        if order is None:
            # default is hill system
            if "C" in elements.keys():
                order = self._set_element_order_in_formula(["C", "H"])
            else:
                order = self._set_element_order_in_formula()
        else:
            order = self._set_element_order_in_formula(order)

        # Generate chemical formula
        chemical_formula = ""
        for element in order:
            if element in elements.keys() and (elements[element] is not None or elements[element] == 0):
                if elements[element] == 1:
                    chemical_formula += element
                else:
                    chemical_formula += element + str(elements[element])

        return chemical_formula

    @staticmethod
    def _set_element_order_in_formula(order: list[str] = None) -> list[str]:
        """Given a chemical order of element_symbols, validate it/ and complete it. Default is alphabetically."""
        base = periodic_table.symbols_alphabetical_order()
        if order is None:
            return base

        for k in reversed(order):
            if k in base:
                base.remove(k)
                base.insert(0, k)
            else:
                raise MolecularFormulaError(f"Invalid symbol in 'order' for chemical formula. Invalid: {k}")

        return base


def local_run():
    cf = MolecularFormula("C1H3OH63Cr2CCCOOO")
    print(cf)


if __name__ == '__main__':
    local_run()



"""

The following contains the following data structures:
* Atom
* Bond
* Branch
* Bond descriptor
* Bond descriptor atom
* Stochastic fragment
* Stochastic object
* BigSMILES

"""

from __future__ import annotations
import logging

import bigsmiles.errors as errors
import bigsmiles.reference_data.chemical_data as chemical_data
from bigsmiles.config import Config

_conjugated_warning = True


class Atom:
    """
    this class represents an atom
    """
    __slots__ = ["id_", "symbol", "isotope", "stereo", "hydrogens", "charge", "class_", "organic",
                 "aromatic", "valence", "possible_valence", "_default_valence", "_valene_warning_raised", "_bonds",
                 "parent", "__dict__"]
    _eq_attr = ("id_", "symbol", "isotope", "stereo", "hydrogens", "charge", "valence", "aromatic")

    def __init__(self,
                 id_: int,
                 symbol: str,
                 isotope: int | None = None,
                 stereo: str | None = None,
                 hydrogens: int | None = None,
                 charge: int = 0,
                 valence: int | float | None = None,
                 class_: int | None = None,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs
                 ):
        """
        Attributes
        ----------
        id_: int
            id of atom (id is limited to atoms). Range: [1, inf]
        symbol: str
            symbol symbol (e.g., H, C, O, Zn)
        isotope: int | None
            isotope (e.g., [13C])
        stereo: str | None
            stereochemistry [None, "@", "@@"] (None means not defined)
        hydrogens: int | None
            number of explict hydrogens  (e.g., [CH2])
        charge: int
            symbol charge  (e.g., [Fe+3])
        valence: int
            The capacity to form bonds with other atoms
        class_: int | None
            index of class (e.g., [C:1] class_ = 1)
        parent: BigSMILES | Branch | StochasticFragment | None
            the owner of the atom
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.id_ = id_
        self.symbol = symbol.capitalize()
        self.isotope = isotope
        self.stereo = stereo
        self.hydrogens = hydrogens
        self.charge = charge
        self.class_ = class_
        self.organic = True if self.symbol in chemical_data.organic_elements else False
        # TODO: calculate aromatic
        self.aromatic = True if symbol in chemical_data.aromatic_elements else False
        self.possible_valence: tuple[int] = chemical_data.atom_valences[self.symbol]
        if valence is None:
            self._default_valence: bool = True
            self.valence = self.possible_valence[0]
        else:
            self._default_valence: bool = False
            self.valence = valence

        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

        self._bonds = []
        self._valene_warning_raised = False

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(Atom,))

    def __eq__(self, other: Atom):
        """
        equality is based on the following parameters:

        * id_
        * symbol
        * isotope
        * stereo
        * hydrogens
        * charge
        * valence
        * aromatic
        * bonds ~ bond.id_
        """
        if not isinstance(other, Atom):
            return False

        # check attributes
        for attr in self._eq_attr:
            if getattr(self, attr) != getattr(other, attr):
                return False

        # check atoms
        for bond, other_bond in zip(self.bonds, other.bonds):
            if bond.id_ != other_bond.id_:
                return False

        return True

    # def __hash__(self):
    #     NotImplemented and shouldn't according to Python docs because bond is mutable. If the Atom is hashed, a bond
    #     add; then the hash would be wrong. Instead, use '.id_' for this purpose.
    #     Docs: https://docs.python.org/3/reference/datamodel.html#object.__hash__
    #

    @property
    def bonds(self) -> list[Bond]:
        """ list of bonds (excludes implicit hydrogens) """
        return self._bonds

    @bonds.setter
    def bonds(self, bonds: list[Bond]):
        self._bonds = bonds
        # Validation for available valence
        if self.number_of_bonds > self.bond_capacity:
            if not self._increase_valence(self.number_of_bonds - self.bond_capacity):
                logging.error(f"Too many bonds trying to be made. {str(self)}")

    def delete_bond(self, delete_bond: Bond | list[Bond]):
        """
        delete bond or bonds

        Parameters
        ----------
        delete_bond: Bond | list[Bond]

        """
        if not isinstance(delete_bond, list):
            delete_bond = [delete_bond]

        for bond in delete_bond:
            try:
                self.bonds.remove(bond)
            except ValueError:
                raise ValueError("Bond not connect to atom so it can't be deleted."
                                 f"\n Atom:{self.details}\nBond attempting to remove: {bond.details}")

    @property
    def implicit_hydrogens(self) -> float:
        """ number of implicit hydrogens. (zero if explict hydrogens have been specified) """
        return self.bonds_available if self.hydrogens is None else 0

    @property
    def bond_capacity(self) -> float:
        """ total capacity of the atom account for charge """
        return self.valence + self.charge

    @property
    def number_of_bonds(self) -> float:
        """ number of bonds already formed (not including implicit hydrogens; but including explicit hydrogens) """
        num_bonds = sum((bond.bond_order for bond in self.bonds))
        if self.hydrogens is not None:
            num_bonds += self.hydrogens
        return num_bonds

    @property
    def bonds_available(self) -> float:
        """ number bonds that remain open for bonding """
        bonds_available = self.bond_capacity - self.number_of_bonds
        if bonds_available < 0:
            return 0
        return bonds_available

    @property
    def full_valence(self) -> bool:
        """ returns true if the atom valence is full """
        if self.valence - self.number_of_bonds - self.implicit_hydrogens - abs(self.charge) == 0:
            return True

        return False

    @property
    def ring_indexes(self) -> list[int]:
        """ list of ring index """
        ring_index = []
        for bond in self.bonds:
            if bond.ring_id is not None:
                ring_index.append(bond.ring_id)

        return ring_index

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        for attr in self._eq_attr:
            text += f"{attr}: {getattr(self, attr)}, "
        return text[:-2] + "}"

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = self._to_string(show_hydrogens, show_atom_index, show_repr)
        if not self.full_valence and not self._valene_warning_raised:
            self._valene_warning_raised = True
            logging.warning(f"Incomplete valence detected on atom: {text}")

        return text

    def _to_string(self, show_hydrogens: bool = False, show_atom_index: bool = True, show_repr: tuple | None = None)\
            -> str:
        text = ""
        bracket_flag = False

        if self.symbol == "H":
            bracket_flag = True

        if self.isotope is not None:
            text += str(self.isotope)
            bracket_flag = True

        if self.aromatic:
            text += self.symbol.lower()
        else:
            text += self.symbol

        if self.stereo:
            text += self.stereo
            bracket_flag = True

        if show_hydrogens or self.hydrogens is not None:
            if self.hydrogens is not None:
                if self.hydrogens == 1:
                    text += f"H"
                elif self.hydrogens > 1:
                    text += f"H{self.hydrogens}"

                bracket_flag = True
            else:
                if self.implicit_hydrogens == 1:
                    text += f"H"
                    bracket_flag = True
                if self.implicit_hydrogens > 1:
                    text += f"H{self.implicit_hydrogens}"
                    bracket_flag = True

        if self.charge > 0:
            text += f"+{self.charge if self.charge > 1 else ''}"
            bracket_flag = True
        elif self.charge < 0:
            text += f"-{self.charge if self.charge > 1 else ''}"
            bracket_flag = True

        if self.class_ is not None and show_atom_index:
            text += ":" + str(self.class_)

        if bracket_flag:
            text = "[" + text + "]"

        if self.ring_indexes:
            # get ring ids
            rings = []
            for bond in self.bonds:
                if bond.ring_id is not None:
                    rings.append([bond.ring_id, bond])

            # sort to numerical order
            rings.sort(key=lambda x: x[0])
            for ring_id, bond in rings:
                ring_text = ""
                if bond.symbol != ":":  # don't show aromatic bonds in ring index
                    if not self.root.contains_stochastic_object and not Config.show_multi_bonds_on_both_ring_index \
                            and bond.atom2 == self:
                        pass
                    else:
                        ring_text += bond.symbol
                if ring_id > 9:
                    ring_text += "%"

                text += ring_text + str(ring_id)

        if show_repr is not None and Atom in show_repr:
            text += "{" + str(self.id_) + "}"

        return text

    def _increase_valence(self, requested_valence_increase: float = 0) -> bool:
        """ If default valence, try to increase when an additional bond added that needs it. And does the increase!! """
        if not self._default_valence:
            return False

        for valence in self.possible_valence:
            if valence > self.valence:
                old_valence = self.valence
                self.valence = valence
                if self.bonds_available < requested_valence_increase:
                    # increase was not enough for requested valence increase; set valence back to original value
                    self.valence = old_valence
                    continue

                return True

        return False

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root


class Bond:
    """
    this class represents a bond
    """
    __slots__ = ["id_", "symbol", "atom1", "atom2", "ring_id", "parent", "__dict__", "double_bond_stereo"]
    _eq_attr = ("id_", "symbol", "ring_id") # "double_bond_ez"

    def __init__(self,
                 id_: int,
                 symbol: str,
                 atom1: Atom | BondDescriptorAtom | StochasticObject,
                 atom2: Atom | BondDescriptorAtom | StochasticObject | None = None,
                 ring_id: int | None = None,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs
                 ):
        """

        Parameters
        ----------
        id_: int
            id of bond (id is limited to bonds). Range: [1, inf]
        symbol: str
            bond symbol (e.g. '', '=', '#')
        atom1: Atom | BondDescriptorAtom | StochasticObject
            atom that the bond starts at
        atom2: Atom | BondDescriptorAtom | StochasticObject | None
            atom that the bond ends at
        ring_id: int | None
            if the bond is closing a ring, this is the index for that ring
        parent: BigSMILES | Branch | StochasticFragment | None
            the owner of the bond
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.id_ = id_
        self.symbol = symbol
        self.atom1 = atom1
        self.atom2 = atom2
        self.ring_id = ring_id
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(Bond,), skip_color=True)

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = ""
        if self.symbol != ":" or (self.symbol == ":" and Config.show_aromatic_bond):
            text += Config.add_color(self.symbol, 'Blue', skip_color)

        if show_repr is not None and Bond in show_repr:
            text += "{" + str(self.id_) + "}"

        return text

    def __iter__(self):
        """ iterates over Atoms """
        return iter((self.atom1, self.atom2))

    def __reversed__(self):
        return iter((self.atom2, self.atom1))

    def __eq__(self, other: Bond):
        """
        equality is based on the following parameters:

        * id_
        * symbol
        * ring_id
        * atom1.id_
        * atom2.id_
         """
        if not isinstance(other, Bond):
            return False

        # check attributes
        for attr in self._eq_attr:
            if getattr(self, attr) != getattr(other, attr):
                return False

        # check atoms
        if self.atom1.id_ != other.atom1.id_ or self.atom2.id_ != other.atom2.id_:
            return False

        return True

    def delete(self):
        """ deletes bond and removes it from atoms and root """
        self.atom1.delete_bond(self)
        self.atom2.delete_bond(self)
        self.parent.nodes.remove(self)
        if self.parent is not self.root:
            self.root.bonds.remove(self)
        del self

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        for attr in self._eq_attr:
            text += f"{attr}: {getattr(self, attr)}, "
        text += f"bond_order: {self.bond_order}, "
        text += "atoms: " + repr(self.atom1) + " <--> " + repr(self.atom2)
        return text + "}"

    @property
    def bond_order(self) -> int:
        """  number of covalent bonds represented """
        return chemical_data.bond_mapping[self.symbol]

    @bond_order.setter
    def bond_order(self, count: int):
        for k, v in chemical_data.bond_mapping.items():
            if v == count:
                self.symbol = k
                return

        raise ValueError(f"Invalid bond_order. \nGiven: {count} \n Acceptable: {chemical_data.bond_mapping}")

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    # @property
    # def double_bond_ez(self) -> str | None:
    #     """
    #     returns labels for geometric isomers of alkenes ["E", "Z", None]
    #
    #     ??? info "E/Z more information"
    #
    #         ::: bigsmiles.data_structures.stereo_rules.get_double_bond_ez
    #
    #     """
    #     if self.symbol != "=":
    #         return None
    #
    #     from bigsmiles.data_structures.stereo_rules import get_double_bond_ez
    #     return get_double_bond_ez(self)

    @property
    def aromatic(self) -> bool:
        """ Limited accuracy """
        if self.symbol == ":":
            return True
        return False

    # @property
    # def conjugated(self) -> bool:
    #     # TODO: small molecules and bridge across stochastic object and stochastic fragments
    #     # only for
    #
    #     global _conjugated_warning  # so it is only raised once
    #     if _conjugated_warning and self.root.contains_stochastic_object:
    #
    #         logging.warning("May be wrong for bonds bridging stochastic objects and stochastic fragments.")
    #         _conjugated_warning = False
    #
    #     if self.bond_order <= 1:
    #         return False
    #
    #     if self.aromatic:
    #         return True
    #
    #     # a depth of two graph traversal looking for a double or triple bond
    #     for atom in self:
    #         for bond in atom.bonds:
    #             if bond.id_ == self.id_:
    #                 continue  # skip self
    #             for atom_ in bond:
    #                 if atom_.id_ == atom.id_:
    #                     continue  # skip if past atom
    #                 for bond_ in atom_.bonds:
    #                     if bond_.bond_order > 1:
    #                         return True
    #
    #     return False


class BondDescriptor:
    """
    this class represents a bonding descriptor
    """
    __slots__ = ["parent", "descriptor", "index_", "_instances", "_instances_up_to_date", "bond_symbol", "__dict__"]
    _eq_attr = ("descriptor", "index_", "bond_symbol")

    def __init__(self,
                 parent: StochasticObject,
                 descriptor: str,
                 index_: int,
                 bond_symbol: str | None,
                 **kwargs
                 ):
        """

        Parameters
        ----------
        parent: StochasticObject
            the owner of the bonding descriptor
        descriptor: str
            descriptor symbol [$, >, <]
        index_: int
            descriptor index [1, inf]
        bond_symbol: str | None
            bond symbol
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.parent = parent
        self.descriptor = descriptor
        self.index_ = index_

        self._instances_up_to_date: bool = True
        self._instances: list[BondDescriptorAtom] = []
        self.bond_symbol: str | None = bond_symbol

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(BondDescriptor,), skip_color=True)

    def __eq__(self, other: BondDescriptor):
        """
        equality is based on the following parameters:

        * descriptor
        * index_
        * bond_symbol
        """
        if not isinstance(other, BondDescriptor):
            return False

        for attr in self._eq_attr:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = "[" + self.descriptor
        if self.index_ == 1 and not Config.show_bond_descriptor_one_index:
            text += ""
        else:
            text += str(self.index_)
        text += "]"

        return text

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        for attr in self._eq_attr:
            text += f"{attr}: {getattr(self, attr)}, "
        return text[:-1] + "}"

    @property
    def symbol(self) -> str:
        """ descriptor + index """
        return self.descriptor + str(self.index_)

    def is_pair(self, bd: BondDescriptor) -> bool:
        """ Returns true if symbols are <> and index match. """
        if self.index_ == bd.index_:
            if (self.descriptor == "<" and bd.descriptor == ">") or \
                    (self.descriptor == ">" and bd.descriptor == "<"):
                return True

        return False

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    @property
    def instances(self) -> list[BondDescriptorAtom]:
        """ list of BondDescriptorAtom that are based on this bonding descriptor """
        return self._instances

    @instances.setter
    def instances(self, instances: list[BondDescriptor]):
        self._instances = instances
        self._instances_up_to_date = False

    @property
    def bond_order(self) -> int | None:
        """ number of covalent bonds represented """
        return chemical_data.bond_mapping[self.bond_symbol]

    @property
    def implicit(self) -> bool:
        """ returns true is it is an implicit bonding descriptor """
        if self.descriptor == "":
            return True
        return False

    @property
    def aromatic(self) -> bool:
        """ limited accuracy """
        return self.bond_symbol == ":"


class BondDescriptorAtom:
    """
    this class represents a bonding descriptor and has the ability to bond
    it should be thought of as an atom with only on bonding site that can be any bond (single, double triple)
    """
    __slots__ = ["descriptor", "id_", "_bond", "__dict__", "parent"]

    def __init__(self,
                 id_: int,
                 bond_descriptor: BondDescriptor,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs
                 ):
        """
        Parameters
        ----------
        id_: int
            id of bonding descriptor atom (id is limited to bonding descriptor atoms). Range: [1, inf]
        bond_descriptor: BondDescriptor
            bonding descriptor that this atom represents
        parent: BigSMILES | Branch | StochasticFragment | None
            the owner of the bonding descriptor atom
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.descriptor = bond_descriptor
        bond_descriptor.instances += [self]
        self.id_ = id_
        self._bond = None  # Added after construction
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(BondDescriptorAtom,), skip_color=True)

    def __eq__(self, other: BondDescriptorAtom):
        """
        equality is based on the following parameters:

        * id_
        * descriptor
        * bond.id_
        """
        if not isinstance(other, BondDescriptorAtom):
            return False

        if self.id_ != other.id_:
            return False

        if self.descriptor != other.descriptor:
            return False

        if self.bond is None:
            if other.bond is not None:
                return False
        else:
            if other.bond is None or self.id_ != other.id_:
                return False

        return True

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = Config.add_color(self.descriptor.to_string(show_hydrogens, show_repr), 'Green', skip_color)

        if show_repr is not None and BondDescriptorAtom in show_repr:
            text += "{" + str(self.id_) + "}"

        return text

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        text += f"id_: {self.id_}"
        return text + "}"

    @property
    def bond(self) -> Bond:
        """ bond to bonding descriptor atom """
        return self._bond

    @bond.setter
    def bond(self, bond: Bond):
        if self.bond is not None:
            raise errors.ConstructorError(f"Trying to make a bond to {self} when it already has a bond.")
        self._bond = bond

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    @property
    def aromatic(self) -> bool:
        """ limited accuracy """
        return self.descriptor.aromatic


class Branch:
    """
    this class represents a branch
    """
    __slots__ = ["nodes", "id_", "parent", "__dict__"]

    def __init__(self, id_: int, parent: BigSMILES | StochasticFragment | Branch, **kwargs):
        """
        Parameters
        ----------
        id_: int
            id of bond (id is limited to bonds). Range: [1, inf]
        parent: BigSMILES | StochasticFragment | Branch
            owner of the branch
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.nodes: list[Atom | Bond | Branch | StochasticObject | BondDescriptorAtom] = []
        self.id_ = id_
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(Branch,), skip_color=True)

    def __eq__(self, other: Branch):
        """ equality is determining all nodes and rings are equal """
        if not isinstance(other, Branch):
            return False

        for node, other_node in zip(self.nodes, other.nodes):
            if node != other_node:
                return False

        return True

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = "".join(node.to_string(show_hydrogens, show_atom_index, show_repr, skip_color) for node in self.nodes)

        if show_repr is not None and Branch in show_repr:
            text += "{" + str(self.id_) + "}"

        return "(" + text + ")"

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        text += f"id_: {self.id_}, "
        text += f"num_nodes: {len(self.nodes)}"
        return text + "}"

    @property
    def in_stochastic_object(self) -> bool:
        """ returns that if on the tree up to the root there is a stochastic object parent """
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    def _get_id(self, node_type) -> int:
        return self.root._get_id(node_type)  # noqa


class StochasticFragment:
    """
    this class represents a stochastic fragment.

    stochastic fragments are only found within stochastic objects
    """
    __slots__ = ["nodes", "id_", "parent", "bonding_descriptors", "rings", "__dict__"]

    def __init__(self, id_: int, parent: StochasticObject, **kwargs):
        """
        Parameters
        ----------
        id_: int
            id of bond (id is limited to bonds). Range: [1, inf]
        parent
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.id_ = id_
        self.parent = parent
        self.nodes: list[Atom | Bond | BondDescriptorAtom | Branch | StochasticObject] = []
        self.rings: list[Bond] = []
        self.bonding_descriptors: list[BondDescriptor] = []

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(StochasticFragment,), skip_color=True)

    def __eq__(self, other: StochasticFragment):
        """ equality is determining all nodes and rings are equal """
        if not isinstance(other, StochasticFragment):
            return False

        for node, other_node in zip(self.nodes, other.nodes):
            if node != other_node:
                return False

        for ring, other_ring in zip(self.rings, other.rings):
            if ring != other_ring:
                return False

        return True

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = "".join(node.to_string(show_hydrogens, show_atom_index, show_repr, skip_color) for node in self.nodes)

        if show_repr is not None and StochasticFragment in show_repr:
            text += "{" + str(self.id_) + "}"

        return text

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        text += f"id_: {self.id_}, "
        text += f"num_nodes: {len(self.nodes)}"
        return text + "}"

    @property
    def in_stochastic_object(self) -> bool:
        """ returns that if on the tree up to the root there is a stochastic object parent """
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    def _get_id(self, node_type) -> int:
        return self.root._get_id(node_type)  # noqa


class StochasticObject:
    """
    this class is represents a stochastic object
    """
    __slots__ = ["nodes", "bonding_descriptors", "id_", "parent", "bd_left", "bd_right",
                 "_bond_left", "_bond_right", "__dict__"]

    def __init__(self, id_: int, parent: BigSMILES | StochasticFragment | Branch, **kwargs):
        """
        Parameters
        ----------
        id_: int
            id of bond (id is limited to bonds). Range: [1, inf]
        parent: BigSMILES | StochasticFragment | Branch
            the owner of the atom
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.nodes: list[StochasticFragment] = []
        self.bonding_descriptors: list[BondDescriptor] = []
        self.bd_left: BondDescriptor | None = None
        self.bd_right: BondDescriptor | None = None
        self.id_ = id_
        self._bond_left: Bond | None = None
        self._bond_right: Bond | None = None
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(StochasticObject,), skip_color=True)

    def __eq__(self, other: StochasticObject):
        """ equality is determining all nodes and bonding descriptors are equal """
        if not isinstance(other, StochasticObject):
            return False

        for node, other_node in zip(self.nodes, other.nodes):
            if node != other_node:
                return False

        if self.bd_left != other.bd_left:
            return False

        if self.bd_right != other.bd_right:
            return False

        return True

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        text += f"id_: {self.id_}, "
        text += f"num_nodes: {len(self.nodes)}"
        return text + "}"

    @property
    def bond_left(self) -> Bond | None:
        """ bond leaving the left of the stochastic object """
        return self._bond_left

    @bond_left.setter
    def bond_left(self, bond: Bond):
        if self._bond_left is not None:
            raise errors.ConstructorError(f"Trying to make a bond to {self} when it already has a bond.")
        self._bond_left = bond

    @property
    def bond_right(self) -> Bond | None:
        """ bond leaving the right of the stochastic object """
        return self._bond_right

    @bond_right.setter
    def bond_right(self, bond: Bond):
        if self._bond_right is not None:
            raise errors.ConstructorError(f"Trying to make a bond to {self} when it already has a bond.")
        self._bond_right = bond

    @property
    def aromatic(self) -> bool:
        """ limited accuracy """
        if self.bd_left.bond_order == 1.5:
            # not checking right bond descriptor, maybe we should? but during construction it is not defined
            return True
        return False

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        text = ""
        text += Config.add_color("{", "Red", skip_color)
        if self.bd_left is not None:
            text += self.bd_left.to_string(show_hydrogens, show_atom_index, show_repr, skip_color)
        text += ",".join(node.to_string(show_hydrogens, show_atom_index, show_repr, skip_color) for node in self.nodes)
        if self.bd_right is not None:
            text += self.bd_right.to_string(show_hydrogens, show_atom_index, show_repr, skip_color)
        text += Config.add_color("}", "Red", skip_color)
        if self.bond_right is not None and self.bond_right.ring_id is not None:
            if self.bond_right.symbol != ":":
                text += self.bond_right.symbol
            text += str(self.bond_right.ring_id)

        if show_repr is not None and StochasticObject in show_repr:
            text += "{" + str(self.id_) + "}"

        return text

    @property
    def implicit_endgroups(self) -> bool:
        """ Returns true if one or more are implicit """
        if (self.bd_left is not None and self.bd_left.implicit) or \
                (self.bd_right is not None and self.bd_right.implicit):
            return True

        return False

    @property
    def in_stochastic_object(self) -> bool:
        """ returns True as it is the stochastic object """
        return True

    @property
    def root(self) -> BigSMILES:
        """ the owner at the top of the parent tree """
        return self.parent.root

    @property
    def bonds(self) -> list[Bond]:
        """ bonds entering the left and right """
        bonds = []
        if self.bond_left is not None:
            bonds.append(self.bond_left)
        if self.bond_right is not None:
            bonds.append(self.bond_right)
        return bonds

    def _get_id(self, node_type) -> int:
        return self.root._get_id(node_type)  # noqa


def contains_stochastic_object(nodes: list[Atom, Bond, Branch, StochasticObject]):
    """ recursive search for a stochastic_object. """
    for node in nodes:
        if isinstance(node, StochasticObject):
            return True
        if isinstance(node, Branch):
            result = contains_stochastic_object(node.nodes)
            if result:
                return True

    return False


class BigSMILES:
    """
    this class is used to represent a BigSMILES.

    BigSMILES is a superset of SMILES; so any this class handles SMILES as well

    Examples
    --------
    bigsmiles("C1CCCCC1")

    bigsmiles("CC(CC){[>][<]CC(c1ccccc1)[>][<]}[H]")

    """
    __slots__ = ["nodes", "atoms", "bonds", "rings", "__dict__"]

    def __init__(self, text: str | None = None, **kwargs):
        """
        Parameters
        ----------
        text: str | None
            BigSMILES or SMILES string. String will be processed automatically. (pass 'None' if using 'constructor')
        kwargs:
            any additional keyword arguments are accepted and set as additional attributes
        """
        self.nodes: list[Atom | Bond | StochasticObject | Branch] = []
        self.atoms: list[Atom | BondDescriptorAtom] = []  # includes atoms in sub-objects
        self.bonds: list[Bond] = []  # includes bonds in sub-objects
        self.rings: list[Bond] = []  # does not include rings in sub-objects

        self._ids_ = {Atom: 0, Bond: 0, BondDescriptorAtom: 0, StochasticFragment: 0, StochasticObject: 0, Branch: 0}

        # parse input string
        if text:
            # import here to avoid circular imports
            from bigsmiles.constructors.constructor_tokens import parse_bigsmiles_str
            parse_bigsmiles_str(text, self)

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(show_repr=(Atom, Bond, Branch, StochasticFragment, StochasticObject, BondDescriptorAtom),
                              skip_color=True)

    def __getitem__(self, obj: int | slice) -> Atom | list[Atom]:
        """ get items by node index or slice """
        return self.nodes[obj]

    def __iter__(self):
        """ iterator over nodes """
        return iter(self.nodes)

    def __bool__(self):
        """ returns True if BigSMILES has nodes """
        if self.nodes:
            return True
        else:
            return False

    def __eq__(self, other: BigSMILES):
        """ equality is determining all nodes and rings are equal """
        if not isinstance(other, BigSMILES):
            return False

        for node, other_node in zip(self.nodes, other.nodes):
            if node != other_node:
                return False

        for ring, other_ring in zip(self.rings, other.rings):
            if ring != other_ring:
                return False

        return True

    def to_string(self,
                  show_hydrogens: bool = False,
                  show_atom_index: bool = True,
                  show_repr: tuple | None = None,
                  skip_color: bool = False
                  ) -> str:
        return "".join(node.to_string(show_hydrogens, show_atom_index, show_repr, skip_color) for node in self.nodes)

    @property
    def details(self) -> str:
        """ long string representation """
        text = str(self) + "  {"
        text += f"num_nodes: {len(self.nodes)}, "
        text += f"num_atoms: {len(self.atoms)}, "
        text += f"num_bonds: {len(self.bonds)}"
        return text + "}"

    @property
    def in_stochastic_object(self) -> bool:
        """ returns False as BigSMILES is always a root node and can't be contained in anything """
        return False

    @property
    def contains_stochastic_object(self) -> bool:
        """ returns True if a stochastic object is present in the BigSMILES (anywhere) """
        return contains_stochastic_object(self.nodes)

    @property
    def has_disconnect(self) -> bool:
        """ returns True if '.' (disconnect symbol) is present in the BigSMILES (anywhere) """
        for bond in self.bonds:
            if bond.symbol == ".":
                return True
        return False

    @property
    def root(self) -> BigSMILES:
        """ root will return itself as BigSMILES object is always the root node """
        return self

    def _get_id(self, node_type) -> int:
        self._ids_[node_type] += 1
        return self._ids_[node_type]


# types  (for python >3.10)
# has_node_attr = BigSMILES | Branch | StochasticObject | StochasticFragment
# has_ring_attr = BigSMILES | StochasticFragment
# has_parent_attr = Branch | StochasticObject | StochasticFragment | Bond | BondDescriptorAtom | Atom
# has_root_attr = BigSMILES | has_parent_attr

# types (for python <=3.9)
import typing

bigsmiles_obj = [Atom, Bond, Branch, BondDescriptorAtom, StochasticObject, StochasticFragment, BigSMILES]
has_node_attr = typing.Union[BigSMILES, Branch, StochasticObject, StochasticFragment]
has_ring_attr = typing.Union[BigSMILES, StochasticObject]
has_parent_attr = typing.Union[Branch, StochasticObject, StochasticFragment, Bond, BondDescriptorAtom, Atom]
has_root_attr = typing.Union[BigSMILES, has_parent_attr]




class TerminalColors:
    Bold = '\033[1m'
    Underline = '\033[4m'
    Italics = '\33[3m'

    Red = '\33[91m'
    Green = '\33[92m'
    Yellow = '\33[93m'
    Blue = '\33[94m'
    Purple = '\33[95m'
    Cyan = '\33[96m'
    White = '\33[97m'
    Black = '\33[90m'

    Clear = '\33[0m'


class Config:
    # Will show colors in terminal when any object is printed
    color_output = False

    # True: [$1]; False: [$]
    show_bond_descriptor_one_index = False

    # True: will add ':' symbol for aromatic_elements bonds, False: hides the ':' in string outputs
    show_aromatic_bond = True

    # for small molecules only. True: 'C=1CCCCC=1'  False: 'C=1CCCCC1'
    show_multi_bonds_on_both_ring_index = False

    @classmethod
    def add_color(cls, text: str, color: str, skip_color: bool = False) -> str:
        if cls.color_output and not skip_color:
            return getattr(TerminalColors, color) + text + TerminalColors.Clear

        return text



"""

Code for turing a python class into a terminal printable tree.

"""
from __future__ import annotations

from bigsmiles.data_structures.bigsmiles import Atom, Bond, BondDescriptorAtom, Branch, StochasticFragment, \
    StochasticObject, BigSMILES


class TreeConfig:
    """
    this class can be used to configure the tree parsing output

    Parameters
    ----------
    tree_symbol_options: dict
        symbols to build the tree. various ascii options are available
    symbols: str
        select from 'tree_symbol_options' what ascii symbols you want to use

    Notes
    -----
    * color can be added to the output with through the main configuration class

    """
    tree_symbol_options = {
        'ascii': ('|', '|-- ', '+-- '),
        'ascii-ex': ('\u2502', '\u251c\u2500\u2500', '\u2514\u2500\u2500'),
        'ascii-exr': ('\u2502', '\u251c\u2500\u2500', '\u2570\u2500\u2500'),
        'ascii-em': ('\u2551', '\u2560\u2550\u2550', '\u255a\u2550\u2550'),
        'ascii-emv': ('\u2551', '\u255f\u2500\u2500', '\u2559\u2500\u2500'),
        'ascii-emh': ('\u2502', '\u255e\u2550\u2550', '\u2558\u2550\u2550'),
    }
    symbols = 'ascii-ex'  # some computers may need to show different symbols
    tab = "    "
    space = " "
    printed_classes = {
        Atom: True,
        Bond: True,
        BondDescriptorAtom: True,
        Branch: False,
        StochasticFragment: False,
        StochasticObject: False,
        BigSMILES: False
    }

    @classmethod
    def vertical_line(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][0]

    @classmethod
    def tee(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][1]

    @classmethod
    def corner(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][2]


def to_repr(obj) -> str:
    return repr(obj)


def to_str(obj) -> str:
    return str(obj)


def tree_to_string(bigsmiles_: BigSMILES, show_object_labels: bool = True, print_repr: bool = False) -> str:
    """
    Creates a string representation of a tree for printing.

    Parameters
    ----------
    bigsmiles_: BigSMILES
        Root node
    show_object_labels: bool
        show object labels
    print_repr: bool
        print str or print repr of objects
        True: repr() is called
        False: str() is called

    Returns
    -------
    text: str
        String with the structure of a tree.

    """
    text = ""

    # add root node
    if show_object_labels:
        text = type(bigsmiles_).__name__ + ": "

    text += str(bigsmiles_)

    # add main part of tree
    text += tree_to_string_loop(bigsmiles_.nodes, [], show_object_labels, to_repr if print_repr else to_str)

    return text


def tree_to_string_loop(
        nodes: list,
        spacers: list[bool],
        show_object_label: bool,
        func: callable
) -> str:
    """
    Creates the main body of the tree

    Parameters
    ----------
    nodes: list
        nodes
    spacers: list[bool]
        list records weather a continuation line (TreeConfig.vertical_line) is needed in the tree
        True: No line
        False: Add line
    show_object_label: bool
        show object/class type.__name__
    func: callable
        calls str() or repr()

    Returns
    -------
    text: str

    """
    text = ""

    for node in nodes:
        text += create_row(node, spacers, show_object_label, func, node == nodes[-1])

        # if interior node, do recursion
        if hasattr(node, "nodes"):
            text += tree_to_string_loop(
                node.nodes,
                spacers + [node == nodes[-1]],
                show_object_label,
                func
            )

    return text


def create_row(
        node,
        spacers: list[bool],
        show_object_label: bool,
        func: callable,
        last_node: bool
) -> str:
    """
    Creates a row in the tree string representation.

    Parameters
    ----------
    node: TreeNode | TreeIntermediateNode
        node
    spacers: list[bool]
        list records weather a continuation line (TreeConfig.vertical_line) is needed in the tree
        True: No line
        False: Add line
    show_object_label: bool
        show object/class type.__name__
    func: callable
        calls str() or repr()
    last_node: bool
        last node in branch --> switch tee symbol to corner

    Returns
    -------
    text: str

    """
    line = "\n"

    # add spacer
    for i, spacer in enumerate(spacers):
        if not spacer:
            line += TreeConfig.vertical_line()
        line += TreeConfig.tab

    # add tree connection symbol
    if last_node:
        line += TreeConfig.corner() + TreeConfig.space
    else:
        line += TreeConfig.tee() + TreeConfig.space

    # add class label
    if show_object_label:
        line += type(node).__name__ + ": "

    # add node symbol
    if TreeConfig.printed_classes[type(node)]:
        line += func(node)
    else:
        line += str(node)

    return line



import pkg_resources
__version__ = pkg_resources.require("bigsmiles")[0].version

from bigsmiles.config import Config
from bigsmiles.errors import ERRORS
from bigsmiles.data_structures.bigsmiles import Atom, Bond, Branch, BondDescriptor, BondDescriptorAtom, \
    StochasticObject, StochasticFragment, BigSMILES
from bigsmiles.data_structures.reaction import Reaction
